# kytten/checkbox.py
# Copyrighted (C) 2009 by Conrad "Lynx" Wong

import pyglet
from widgets import Control
from layout import HALIGN_LEFT, HALIGN_RIGHT
from override import KyttenLabel

class Checkbox(Control):
    """
    A two-state checkbox.
    """
    def __init__(self, text="", is_checked=False, id=None,
                 align=HALIGN_RIGHT, padding=4, on_click=None,
                 disabled=False):
        """
        Creates a new checkbox.  The provided text will be used to caption the
        checkbox.

        @param text Label for the checkbox
        @param is_checked True if we should start checked
        @param id ID for value
        @param align HALIGN_RIGHT if label should be right of checkbox,
                     HALIGN_LEFT if label should be left of checkbox
        @param padding Space between checkbox and label
        @param on_click Callback for the checkbox
        @param disabled True if the checkbox should be disabled
        """
        assert align in [HALIGN_LEFT, HALIGN_RIGHT]
        Control.__init__(self, id=id, disabled=disabled)
        self.text = text
        self.is_checked = is_checked
        self.align = align
        self.padding = padding
        self.on_click = on_click
        self.label = None
        self.checkbox = None
        self.highlight = None

    def delete(self):
        """
        Clean up our graphic elements
        """
        Control.delete(self)
        if self.checkbox is not None:
            self.checkbox.delete()
            self.checkbox = None
        if self.label is not None:
            self.label.delete()
            self.label = None
        if self.highlight is not None:
            self.highlight.delete()
            self.highlight = None

    def get_value(self):
        return self.is_checked

    def is_input(self):
        return True

    def layout(self, x, y):
        """
        Places the Checkbox.

        @param x X coordinate of lower left corner
        @param y Y coordinate of lower left corner
        """
        Control.layout(self, x, y)
        if self.align == HALIGN_RIGHT:  # label goes on right
            self.checkbox.update(x, y + self.height/2 - self.checkbox.height/2,
                                 self.checkbox.width, self.checkbox.height)
            self.label.x = x + self.checkbox.width + self.padding
        else: # label goes on left
            self.label.x = x
            self.checkbox.update(x + self.label.content_width + self.padding,
                                 y + self.height/2 - self.checkbox.height/2,
                                 self.checkbox.width, self.checkbox.height)

        if self.highlight is not None:
            self.highlight.update(self.x, self.y, self.width, self.height)

        font = self.label.document.get_font()
        height = font.ascent - font.descent
        self.label.y = y + self.height/2 - height/2 - font.descent

    def on_gain_highlight(self):
        Control.on_gain_highlight(self)
        self.size(self.saved_dialog)
        self.highlight.update(self.x, self.y, self.width, self.height)

    def on_lose_highlight(self):
        Control.on_lose_highlight(self)
        if self.highlight is not None:
            self.highlight.delete()
            self.highlight = None

    def on_mouse_press(self, x, y, button, modifiers):
        if not self.is_disabled():
            self.is_checked = not self.is_checked
            if self.on_click is not None:
                if self.id is not None:
                    self.on_click(self.id, self.is_checked)
                else:
                    self.on_click(self.is_checked)

            # Delete the button to force it to be redrawn
            self.delete()
            self.saved_dialog.set_needs_layout()

    def size(self, dialog):
        """
        Sizes the Checkbox.  If necessary, creates the graphic elements.

        @param dialog Dialog which contains the Checkbox
        """
        if dialog is None:
            return
        Control.size(self, dialog)
        if self.is_checked:
            path = ['checkbox', 'checked']
        else:
            path = ['checkbox', 'unchecked']
        if self.is_disabled():
            color = dialog.theme[path]['disabled_color']
        else:
            color = dialog.theme[path]['gui_color']
        if self.checkbox is None:
            self.checkbox = dialog.theme[path]['image'].generate(
                color,
                dialog.batch, dialog.bg_group)
        if self.highlight is None and self.is_highlight():
            self.highlight = dialog.theme[path]['highlight']['image'].generate(
                    dialog.theme[path]['highlight_color'],
                    dialog.batch,
                    dialog.bg_group)
        if self.label is None:
            self.label = KyttenLabel(self.text,
                font_name=dialog.theme[path]['font'],
                font_size=dialog.theme[path]['font_size'],
                color=color,
                batch=dialog.batch, group=dialog.fg_group)

        # Treat the height of the label as ascent + descent
        font = self.label.document.get_font()
        height = font.ascent - font.descent  # descent is negative
        self.width = self.checkbox.width + self.padding + \
            self.label.content_width
        self.height = max(self.checkbox.height, height)

    def teardown(self):
        self.on_click = None
        Control.teardown(self)


# kytten/menu.py
# Copyrighted (C) 2009 by Conrad "Lynx" Wong

import pyglet

from widgets import Widget, Control
from dialog import Dialog
from frame import Frame
from layout import GetRelativePoint, VerticalLayout
from layout import ANCHOR_CENTER, ANCHOR_TOP_LEFT, ANCHOR_BOTTOM_LEFT
from layout import HALIGN_CENTER
from layout import VALIGN_TOP, VALIGN_CENTER, VALIGN_BOTTOM
from override import KyttenLabel
from scrollable import Scrollable

class MenuOption(Control):
    """
    MenuOption is a choice within a menu.  When selected, it inverts
    (inverted color against text-color background) to indicate that it
    has been chosen.
    """
    def __init__(self, text="", anchor=ANCHOR_CENTER, menu=None,
                 disabled=False):
        Control.__init__(self, disabled=disabled)
        self.text = text
        self.anchor = anchor
        self.menu = menu
        self.label = None
        self.background = None
        self.highlight = None
        self.is_selected = False

    def delete(self):
        if self.label is not None:
            self.label.delete()
            self.label = None
        if self.background is not None:
            self.background.delete()
            self.background = None
        if self.highlight is not None:
            self.highlight.delete()
            self.highlight = None

    def expand(self, width, height):
        self.width = width
        self.height = height

    def is_expandable(self):
        return True

    def layout(self, x, y):
        self.x, self.y = x, y
        if self.background is not None:
            self.background.update(x, y, self.width, self.height)
        if self.highlight is not None:
            self.highlight.update(x, y, self.width, self.height)
        font = self.label.document.get_font()
        height = font.ascent - font.descent
        x, y = GetRelativePoint(self, self.anchor,
                                Widget(self.label.content_width, height),
                                self.anchor, (0, 0))
        self.label.x = x
        self.label.y = y - font.descent

    def on_gain_highlight(self):
        Control.on_gain_highlight(self)
        self.size(self.saved_dialog)  # to set up the highlight
        if self.highlight is not None:
            self.highlight.update(self.x, self.y,
                                  self.menu.width, self.height)

    def on_lose_highlight(self):
        Control.on_lose_highlight(self)
        if self.highlight is not None:
            self.highlight.delete()
            self.highlight = None

    def on_mouse_release(self, x, y, button, modifiers):
        self.menu.select(self.text)

    def select(self):
        if self.is_disabled():
            return  # disabled options can't be selected

        self.is_selected = True
        if self.label is not None:
            self.label.delete()
            self.label = None
        self.saved_dialog.set_needs_layout()

    def size(self, dialog):
        if dialog is None:
            return
        Control.size(self, dialog)
        if self.is_selected:
            path = ['menuoption', 'selection']
        else:
            path = ['menuoption']
        if self.label is None:
            if self.is_disabled():
                color = dialog.theme[path]['disabled_color']
            else:
                color = dialog.theme[path]['text_color']
            self.label = KyttenLabel(self.text,
                color=color,
                font_name=dialog.theme[path]['font'],
                font_size=dialog.theme[path]['font_size'],
                batch=dialog.batch,
                group=dialog.fg_group)
            font = self.label.document.get_font()
            self.width = self.label.content_width
            self.height = font.ascent - font.descent

        if self.background is None:
            if self.is_selected:
                self.background = \
                    dialog.theme[path]['highlight']['image'].generate(
                        dialog.theme[path]['gui_color'],
                        dialog.batch,
                        dialog.bg_group)
        if self.highlight is None:
            if self.is_highlight():
                self.highlight = \
                    dialog.theme[path]['highlight']['image'].generate(
                        dialog.theme[path]['highlight_color'],
                        dialog.batch,
                        dialog.highlight_group)

    def unselect(self):
        self.is_selected = False
        if self.label is not None:
            self.label.delete()
            self.label = None
        if self.background is not None:
            self.background.delete()
            self.background = None
        self.saved_dialog.set_needs_layout()

    def teardown(self):
        self.menu = None
        Control.teardown(self)

class Menu(VerticalLayout):
    """
    Menu is a VerticalLayout of MenuOptions.  Moving the mouse across
    MenuOptions highlights them; clicking one selects it and causes Menu
    to send an on_click event.
    """
    def __init__(self, options=[], align=HALIGN_CENTER, padding=4,
                 on_select=None):
        self.align = align
        menu_options = self._make_options(options)
        self.options = dict(zip(options, menu_options))
        self.on_select = on_select
        self.selected = None
        VerticalLayout.__init__(self, menu_options,
                                align=align, padding=padding)

    def _make_options(self, options):
        menu_options = []
        for option in options:
            if option.startswith('-'):
                disabled = True
                option = option[1:]
            else:
                disabled = False
            menu_options.append(MenuOption(option,
                                           anchor=(VALIGN_CENTER, self.align),
                                           menu=self,
                                           disabled=disabled))
        return menu_options

    def get_value(self):
        return self.selected

    def is_input(self):
        return True

    def select(self, text):
        if not text in self.options:
            return

        if self.selected is not None:
            self.options[self.selected].unselect()
        self.selected = text
        menu_option = self.options[text]
        menu_option.select()

        if self.on_select is not None:
            self.on_select(text)

    def set_options(self, options):
        self.delete()
        self.selected = None
        menu_options = self._make_options(options)
        self.options = dict(zip(options, menu_options))
        self.set(menu_options)
        self.saved_dialog.set_needs_layout()

    def teardown(self):
        self.on_select = None
        VerticalLayout.teardown(self)

class Dropdown(Control):
    def __init__(self, options=[], selected=None, id=None,
                 max_height=400, align=VALIGN_TOP, on_select=None,
                 disabled=False):
        assert options
        Control.__init__(self, id=id, disabled=disabled)
        self.options = options
        self.selected = selected or options[0]
        assert self.selected in self.options
        self.max_height = max_height
        self.align = align
        self.on_select = on_select

        self.field = None
        self.label = None
        self.pulldown_menu = None

    def _delete_pulldown_menu(self):
        if self.pulldown_menu is not None:
            self.pulldown_menu.window.remove_handlers(self.pulldown_menu)
            self.pulldown_menu.teardown()
            self.pulldown_menu = None

    def delete(self):
        if self.field is not None:
            self.field.delete()
            self.field = None
        if self.label is not None:
            self.label.delete()
            self.label = None
        self._delete_pulldown_menu()

    def get_value(self):
        return self.selected

    def is_input(self):
        return True

    def on_mouse_release(self, x, y, button, modifiers):
        if self.is_disabled():
            return

        if self.pulldown_menu is not None:
            self._delete_pulldown_menu()  # if it's already up, close it
            return

        # Setup some callbacks for the dialog
        root = self.saved_dialog.get_root()

        def on_escape(dialog):
            self._delete_pulldown_menu()

        def on_select(choice):
            self.selected = choice
            if self.label is not None:
                self.label.delete()
                self.label = None
            self._delete_pulldown_menu()
            self.saved_dialog.set_needs_layout()

            if self.on_select is not None:
                if self.id is not None:
                    self.on_select(self.id, choice)
                else:
                    self.on_select(choice)

        # We'll need the root window to get window size
        width, height = root.window.get_size()

        # Calculate the anchor point and location for the dialog
        if self.align == VALIGN_TOP:
            # Dropdown is at the top, pulldown appears below it
            anchor = ANCHOR_TOP_LEFT
            x = self.x
            y = -(height - self.y - 1)
        else:
            # Dropdown is at the bottom, pulldown appears above it
            anchor = ANCHOR_BOTTOM_LEFT
            x = self.x
            y = self.y + self.height + 1

        # Now to setup the dialog
        self.pulldown_menu = Dialog(
            Frame(
                Scrollable(Menu(options=self.options, on_select=on_select),
                           height=self.max_height),
                path=['dropdown', 'pulldown']
            ),
            window=root.window, batch=root.batch,
            group=root.root_group.parent, theme=root.theme,
            movable=False, anchor=anchor, offset=(x, y),
            on_escape=on_escape)
        root.window.push_handlers(self.pulldown_menu)

    def layout(self, x, y):
        Control.layout(self, x, y)

        self.field.update(x, y, self.width, self.height)
        x, y, width, height = self.field.get_content_region()

        font = self.label.document.get_font()
        height = font.ascent - font.descent
        self.label.x = x
        self.label.y = y - font.descent

    def set_options(self, options, selected=None):
        self.delete()
        self.options = options
        self.selected = selected or self.options[0]
        self.saved_dialog.set_needs_layout()

    def size(self, dialog):
        if dialog is None:
            return
        Control.size(self, dialog)

        if self.is_disabled():
            color = dialog.theme['dropdown']['disabled_color']
        else:
            color = dialog.theme['dropdown']['gui_color']

        if self.field is None:
            self.field = dialog.theme['dropdown']['image'].generate(
                color,
                dialog.batch, dialog.bg_group)
        if self.label is None:
            self.label = KyttenLabel(self.selected,
                font_name=dialog.theme['dropdown']['font'],
                font_size=dialog.theme['dropdown']['font_size'],
                color=dialog.theme['dropdown']['text_color'],
                batch=dialog.batch, group=dialog.fg_group)
        font = self.label.document.get_font()
        height = font.ascent - font.descent
        self.width, self.height = self.field.get_needed_size(
            self.label.content_width, height)

    def teardown(self):
        self.on_select = False
        self._delete_pulldown_menu()
        Control.teardown(self)


# resizable.py
# Copyrighted (C) 2009 by Conrad "Lynx" Wong

import pyglet
from pyglet import gl
import kytten
from kytten.widgets import Control

class Resizable(Control):
    """
    A resizable box.
    """
    def __init__(self, width, height, disabled=False):
        """
        Blocks occupy a fixed width and height.

        @param width Width
        @param height Height
        """
        Control.__init__(self, width=width, height=height, disabled=disabled)
	self.color = None
        self.vertex_list = None
	self.is_dragging = False

    def _get_controls(self):
	if self.width < 20 or self.height < 20:
	    return [(self, self.x, self.x + self.width + 8,
		           self.y + self.height, self.y - 8)]
	else:
	    return Control._get_controls(self)

    def _get_vertices(self):
        """
        Defines the corners of the square.

        @return An array of coordinates for our indexed vertex list.
        """
        x1, y1 = int(self.x), int(self.y)
        x2, y2 = x1 + int(self.width), y1 + int(self.height)
	if self.width < 20 or self.height < 20: # draw drag point on outside
	    x3, y4 = x2, y1
	else:
	    x3, y4 = x2 - 8, y1 + 8
	x4, y3 = x3 + 8, y4 - 8
        return (16, (x1, y1, x2, y1, x2, y1, x2, y2,  # main square
		     x2, y2, x1, y2, x1, y2, x1, y1,
		     x3, y3, x4, y3, x4, y3, x4, y4,  # drag point
		     x4, y4, x3, y4, x3, y4, x3, y3))

    def delete(self):
        """
        Deletes our vertex list.
        """
        if self.vertex_list is not None:
            self.vertex_list.delete()
            self.vertex_list = None

    def layout(self, x, y):
        """Places the vertex list at the new location.

        @param x X coordinate of our lower left corner
        @param y Y coordinate of our lower left corner
        """
        Control.layout(self, x, y)
	if not self.is_disabled():
	    num_points, self.vertex_list.vertices = self._get_vertices()
	    self.vertex_list.colors = self.color * num_points
	else:
	    self.is_dragging = False

    def on_mouse_drag(self, x, y, dx, dy, buttons, modifiers):
	if not self.is_dragging:
	    return

	self.width = max(int(self.width + dx), 0)
	self.height = max(int(self.height - dy), 0)
	if self.saved_dialog is not None:
	    self.saved_dialog.set_needs_layout()

    def on_mouse_press(self, x, y, button, modifiers):
	self.is_dragging = True

    def on_mouse_release(self, x, y, button, modifiers):
	self.is_dragging = False
	self.corner = None

    def size(self, dialog):
        """Constructs a vertex list to draw a crossed square.

        @param dialog The Dialog within which we are contained
        """
        if dialog is None:
            return
        Control.size(self, dialog)
        if self.vertex_list is None and not self.is_disabled():
	    self.color = dialog.theme['gui_color']
	    num_points, vertices = self._get_vertices()
            self.vertex_list = dialog.batch.add(num_points, gl.GL_LINES,
                dialog.fg_group,
                ('v2i', vertices),
                ('c4B', self.color * num_points))




# kytten/input.py
# Copyrighted (C) 2009 by Conrad "Lynx" Wong

import pyglet
from widgets import Control
from override import KyttenInputLabel

class Input(Control):
    """A text input field."""
    def __init__(self, id=None, text="", length=20, max_length=None, padding=0,
                 on_input=None, disabled=False):
        Control.__init__(self, id=id, disabled=disabled)
        self.text = text
        self.length = length
        self.max_length = max_length
        self.padding = padding
        self.on_input = on_input
        self.document = pyglet.text.document.UnformattedDocument(text)
        self.document_style_set = False
        self.text_layout = None
        self.label = None
        self.caret = None
        self.field = None
        self.highlight = None

    def delete(self):
        Control.delete(self)
        if self.caret is not None:
            self.caret.delete()
            self.caret = None
        if self.text_layout is not None:
            self.document.remove_handlers(self.text_layout)
            self.text_layout.delete()
            self.text_layout = None
        if self.label is not None:
            self.label.delete()
            self.label = None
        if self.field is not None:
            self.field.delete()
            self.field = None
        if self.highlight is not None:
            self.highlight.delete()
            self.highlight = None

    def disable(self):
        Control.disable(self)
        self.document_style_set = False

    def enable(self):
        Control.enable(self)
        self.document_style_set = False

    def get_text(self):
        return self.document.text

    def get_value(self):
        return self.get_text()

    def is_focusable(self):
        return True

    def is_input(self):
        return True

    def layout(self, x, y):
        self.x, self.y = x, y
        self.field.update(x, y, self.width, self.height)
        if self.highlight is not None:
            self.highlight.update(x, y, self.width, self.height)

        x, y, width, height = self.field.get_content_region()
        if self.is_focus():
            self.text_layout.begin_update()
            self.text_layout.x = x + self.padding
            self.text_layout.y = y + self.padding
            self.text_layout.end_update()
        else:
            # Adjust the text for font's descent
            descent = self.document.get_font().descent
            self.label.begin_update()
            self.label.x = x + self.padding
            self.label.y = y + self.padding - descent
            self.label.width = width - self.padding * 2
            self.label.end_update()

    def on_gain_highlight(self):
        Control.on_gain_highlight(self)
        self.set_highlight()

    def on_gain_focus(self):
        Control.on_gain_focus(self)
        self.delete()
        if self.saved_dialog is not None:
            self.size(self.saved_dialog)
            self.layout(self.x, self.y)

    def on_key_press(self, symbol, modifiers):
        return pyglet.event.EVENT_HANDLED

    def on_lose_focus(self):
        Control.on_lose_focus(self)
        self.delete()
        if self.saved_dialog is not None:
            self.size(self.saved_dialog)
            self.layout(self.x, self.y)
        if self.on_input is not None:
            if self.id is not None:
                self.on_input(self.id, self.get_text())
            else:
                self.on_input(self.get_text())

    def on_lose_highlight(self):
        Control.on_lose_highlight(self)
        self.remove_highlight()

    def on_mouse_drag(self, x, y, dx, dy, buttons, modifiers):
        if not self.is_disabled() and self.caret:
            return self.caret.on_mouse_drag(x, y, dx, dy, buttons, modifiers)

    def on_mouse_press(self, x, y, button, modifiers):
        if not self.is_disabled():
            return self.caret.on_mouse_press(x, y, button, modifiers)

    def on_text(self, text):
        if not self.is_disabled() and self.caret:
            self.caret.on_text(text)
            if self.max_length and len(self.document.text) > self.max_length:
                self.document.text = self.document.text[:self.max_length]
                self.caret.mark = self.caret.position = self.max_length
            return pyglet.event.EVENT_HANDLED

    def on_text_motion(self, motion):
        if not self.is_disabled() and self.caret:
            return self.caret.on_text_motion(motion)

    def on_text_motion_select(self, motion):
        if not self.is_disabled() and self.caret:
            return self.caret.on_text_motion_select(motion)

    def remove_highlight(self):
        if not self.is_highlight() and not self.is_focus():
            if self.highlight is not None:
                self.highlight.delete()
                self.highlight = None

    def set_highlight(self):
        path = ['input', 'highlight']
        if self.highlight is None:
            self.highlight = self.saved_dialog.theme[path]['image'].generate(
                color=self.saved_dialog.theme[path]['highlight_color'],
                batch=self.saved_dialog.batch,
                group=self.saved_dialog.highlight_group)
            self.highlight.update(self.x, self.y, self.width, self.height)

    def set_text(self, text):
        self.document.text = text
        if self.caret:
            self.caret.mark = self.caret.position = len(self.document.text)
        elif self.label:
            self.label.text = text

    def size(self, dialog):
        if dialog is None:
            return
        Control.size(self, dialog)

        if self.is_disabled():
            color = dialog.theme['input']['disabled_color']
        else:
            color = dialog.theme['input']['text_color']

        # We set the style once.  We shouldn't have to do so again because
        # it's an UnformattedDocument.
        if not self.document_style_set:
            self.document.set_style(0, len(self.document.text),
                                    dict(color=color,
                                         font_name=dialog.theme['font'],
                                         font_size=dialog.theme['font_size']))
            self.document_style_set = True

        # Calculate the needed size based on the font size
        font = self.document.get_font(0)
        height = font.ascent - font.descent
        glyphs = font.get_glyphs('A_')
        width = max([x.width for x in glyphs])
        needed_width = self.length * width + 2 * self.padding
        needed_height = height + 2 * self.padding

        if self.is_focus():
            if self.text_layout is None:
                self.text_layout = pyglet.text.layout.IncrementalTextLayout(
                    self.document, needed_width, needed_height,
                    multiline=False,
                    batch=dialog.batch, group=dialog.fg_group)
                assert self.caret is None
            assert self.label is None
            if self.caret is None:
                self.caret = pyglet.text.caret.Caret(
                    self.text_layout,
                    color=dialog.theme['input']['gui_color'][0:3])
                self.caret.visible = True
                self.caret.mark = 0
                self.caret.position = len(self.document.text)
        else:
            if self.label is None:
                self.label = KyttenInputLabel(self.document.text,
                                              multiline=False,
                                              width=self.width-self.padding*2,
                                              color=color,
                                              batch=dialog.batch,
                                              group=dialog.fg_group)
            assert self.text_layout is None and self.caret is None
        if self.field is None:
            if self.is_disabled():
                color = dialog.theme['input']['disabled_color']
            else:
                color = dialog.theme['input']['gui_color']
            self.field = dialog.theme['input']['image'].generate(
                color=color,
                batch=dialog.batch,
                group=dialog.bg_group)
        if self.highlight is None and self.is_highlight():
            self.set_highlight()

        self.width, self.height = self.field.get_needed_size(
            needed_width, needed_height)

    def teardown(self):
        self.on_input = False
        Control.teardown(self)


# kytten/scrollable.py
# Copyrighted (C) 2009 by Conrad "Lynx" Wong

import pyglet
from pyglet import gl

from dialog import DialogEventManager
from frame import Wrapper
from scrollbar import HScrollbar, VScrollbar
from widgets import Widget

class ScrollableGroup(pyglet.graphics.Group):
    """
    We restrict what's shown within a Scrollable by performing a scissor
    test.
    """
    def __init__(self, x, y, width, height, parent=None):
        """Create a new ScrollableGroup

        @param x X coordinate of lower left corner
        @param y Y coordinate of lower left corner
        @param width Width of scissored region
        @param height Height of scissored region
        @param parent Parent group
        """
        pyglet.graphics.Group.__init__(self, parent)
        self.x, self.y, self.width, self.height = x, y, width, height
        self.was_scissor_enabled = False

    def set_state(self):
        """
        Enables a scissor test on our region
        """
        gl.glPushAttrib(gl.GL_ENABLE_BIT | gl.GL_TRANSFORM_BIT |
                        gl.GL_CURRENT_BIT)
        self.was_scissor_enabled = gl.glIsEnabled(gl.GL_SCISSOR_TEST)
        gl.glEnable(gl.GL_SCISSOR_TEST)
        gl.glScissor(int(self.x), int(self.y),
                     int(self.width), int(self.height))

    def unset_state(self):
        """
        Disables the scissor test
        """
        if not self.was_scissor_enabled:
            gl.glDisable(gl.GL_SCISSOR_TEST)
        gl.glPopAttrib()

class Scrollable(Wrapper):
    """
    Wraps a layout or widget and limits it to a maximum, or fixed, size.
    If the layout exceeds the viewable limits then it is truncated and
    scrollbars will be displayed so the user can pan around.
    """
    def __init__(self, content=None, width=None, height=None,
                 is_fixed_size=False, always_show_scrollbars=False):
        """
        Creates a new Scrollable.

        @param content The layout or Widget to be scrolled
        @param width Maximum width, or None
        @param height Maximum height, or None
        @param is_fixed_size True if we should always be at maximum size;
                             otherwise we shrink to match our content
        @param always_show_scrollbars True if we should always show scrollbars
        """
        if is_fixed_size:
            assert width is not None and height is not None
        Wrapper.__init__(self, content)
        self.max_width = width
        self.max_height = height
        self.is_fixed_size = is_fixed_size
        self.always_show_scrollbars = always_show_scrollbars
        self.hscrollbar = None
        self.vscrollbar = None
        self.content_width = 0
        self.content_height = 0
        self.content_x = 0
        self.content_y = 0
        self.hscrollbar_height = 0
        self.vscrollbar_width = 0

        # We emulate some aspects of Dialog here.  We cannot just inherit
        # from Dialog because pyglet event handling won't allow keyword
        # arguments to be passed through.
        self.theme = None
        self.batch = None
        self.root_group = None
        self.panel_group = None
        self.bg_group = None
        self.fg_group = None
        self.highlight_group = None
        self.needs_layout = False

    def _get_controls(self):
        """
        We represent ourself as a Control to the Dialog, but we pass through
        the events we receive from Dialog.
        """
        base_controls = Wrapper._get_controls(self)
        controls = []
        our_left = self.content_x
        our_right = our_left + self.content_width
        our_bottom = self.content_y
        our_top = our_bottom + self.content_height
        for control, left, right, top, bottom in base_controls:
            controls.append((control,
                             max(left, our_left),
                             min(right, our_right),
                             min(top, our_top),
                             max(bottom, our_bottom)))
        if self.hscrollbar is not None:
            controls += self.hscrollbar._get_controls()
        if self.vscrollbar is not None:
            controls += self.vscrollbar._get_controls()
        return controls

    def delete(self):
        """
        Delete all graphical elements associated with the Scrollable
        """
        Wrapper.delete(self)
        if self.hscrollbar is not None:
            self.hscrollbar.delete()
            self.hscrollbar = None
        if self.vscrollbar is not None:
            self.vscrollbar.delete()
            self.vscrollbar = None
        self.root_group = None
        self.panel_group = None
        self.bg_group = None
        self.fg_group = None
        self.highlight_group = None

    def ensure_visible(self, control):
        """
        Make sure a control is visible.
        """
        offset_x = 0
        if self.hscrollbar:
            offset_x = self.hscrollbar.get(self.content_width,
                                           self.content.width)
        offset_y = 0
        if self.vscrollbar:
            offset_y = self.content.height - self.content_height - \
                     self.vscrollbar.get(self.content_height,
                                         self.content.height)
        control_left = control.x - self.content_x - offset_x
        control_right = control_left + control.width
        control_bottom = control.y - self.content_y + offset_y
        control_top = control_bottom + control.height
        if self.hscrollbar is not None:
            self.hscrollbar.ensure_visible(control_left, control_right,
                                           max(self.content_width,
                                               self.content.width))
        if self.vscrollbar is not None:
            self.vscrollbar.ensure_visible(control_top, control_bottom,
                                           max(self.content_height,
                                               self.content.height))

    def expand(self, width, height):
        if self.content.is_expandable():
            if self.vscrollbar is not None:
                self.content_width = width - self.vscrollbar_width
            else:
                self.content_width = width
            if self.hscrollbar is not None:
                self.content_height = height - self.hscrollbar_height
            else:
                self.content_height = height
            self.content.expand(max(self.content_width, self.content.width),
                                max(self.content_height, self.content.height))
        self.width, self.height = width, height

    def get_root(self):
        if self.saved_dialog:
            return self.saved_dialog.get_root()
        else:
            return self

    def hit_test(self, x, y):
        """
        We only intercept events for the content region, not for
        our scrollbars.  They can handle themselves!
        """
        return x >= self.content_x and y >= self.content_y and \
               x < self.content_x + self.content_width and \
               y < self.content_y + self.content_height

    def is_expandable(self):
        return True

    def layout(self, x, y):
        """
        Reposition the Scrollable

        @param x X coordinate of lower left corner
        @param y Y coordinate of lower left corner
        """
        self.x, self.y = x, y

        # Work out the adjusted content width and height
        if self.hscrollbar is not None:
            self.hscrollbar.layout(x, y)
            y += self.hscrollbar.height
        if self.vscrollbar is not None:
            self.vscrollbar.layout(
                x + self.content_width, y)

        # Set the scissor group
        self.root_group.x, self.root_group.y = x - 1, y - 1
        self.root_group.width = self.content_width + 1
        self.root_group.height = self.content_height + 1

        # Work out the content layout
        self.content_x, self.content_y = x, y
        left = x
        top = y + self.content_height - self.content.height
        if self.hscrollbar:
            left -= self.hscrollbar.get(self.content_width,
                                        self.content.width)
        if self.vscrollbar:
            top += self.vscrollbar.get(self.content_height,
                                       self.content.height)
        self.content.layout(left, top)

        self.needs_layout = False

    def on_update(self, dt):
        """
        On updates, we redo the layout if scrollbars have changed position

        @param dt Time passed since last update event (in seconds)
        """
        if self.needs_layout:
            width, height = self.width, self.height
            self.size(self.saved_dialog)
            self.expand(width, height)
            self.layout(self.x, self.y)

    def set_needs_layout(self):
        self.needs_layout = True
        if self.saved_dialog is not None:
            self.saved_dialog.set_needs_layout()

    def set_wheel_hint(self, control):
        if self.saved_dialog is not None:
            self.saved_dialog.set_wheel_hint(control)

    def set_wheel_target(self, control):
        if self.saved_dialog is not None:
            self.saved_dialog.set_wheel_target(control)

    def size(self, dialog):
        """
        Recalculate the size of the Scrollable.

        @param dialog Dialog which contains us
        """
        if dialog is None:
            return
        Widget.size(self, dialog)
        if self.is_fixed_size:
            self.width, self.height = self.max_width, self.max_height

        self.hscrollbar_height = \
            dialog.theme['hscrollbar']['left']['image'].height
        self.vscrollbar_width = \
            dialog.theme['vscrollbar']['up']['image'].width

        if self.root_group is None: # do we need to re-clone dialog groups?
            self.theme = dialog.theme
            self.batch = dialog.batch
            self.root_group = ScrollableGroup(0, 0, self.width, self.height,
                                              parent=dialog.fg_group)
            self.panel_group = pyglet.graphics.OrderedGroup(
                0, self.root_group)
            self.bg_group = pyglet.graphics.OrderedGroup(
                1, self.root_group)
            self.fg_group = pyglet.graphics.OrderedGroup(
                2, self.root_group)
            self.highlight_group = pyglet.graphics.OrderedGroup(
                3, self.root_group)
            Wrapper.delete(self)  # force children to abandon old groups

        Wrapper.size(self, self)  # all children are to use our groups

        if self.always_show_scrollbars or \
           (self.max_width and self.width > self.max_width):
            if self.hscrollbar is None:
                self.hscrollbar = HScrollbar(self.max_width)
        else:
            if self.hscrollbar is not None:
                self.hscrollbar.delete()
                self.hscrollbar = None

        if self.always_show_scrollbars or \
           (self.max_height and self.height > self.max_height):
            if self.vscrollbar is None:
                self.vscrollbar = VScrollbar(self.max_height)
        else:
            if self.vscrollbar is not None:
                self.vscrollbar.delete()
                self.vscrollbar = None

        self.width = min(self.max_width or self.width, self.width)
        self.content_width = self.width
        self.height = min(self.max_height or self.height, self.height)
        self.content_height = self.height

        if self.hscrollbar is not None:
            self.hscrollbar.size(dialog)
            self.hscrollbar.set(self.max_width, max(self.content.width,
                                                    self.max_width))
            self.height += self.hscrollbar.height

        if self.vscrollbar is not None:
            self.vscrollbar.size(dialog)
            self.vscrollbar.set(self.max_height, max(self.content.height,
                                                     self.max_height))
            self.width += self.vscrollbar.width



# proxy_dialog.py
# Copyrighted (C) 2009 by Conrad "Lynx" Wong

import pyglet
import kytten

class StateManager():
    """StateManager ensures that only the topmost state is handling events
    for the window at a time.  States are expected to implement these
    functions:
        def on_hide_state(self, window, manager):
	    '''Called when the state is hidden, i.e. another state
	    has been pushed on top.

	    @param window The application window
	    @param manager The StateManager
	    '''

	def on_show_state(self, window, manager):
	    '''Called when the state is visible, i.e. it has just been
	    pushed or a state that was on top has been removed.

	    @param window The application window
	    @param manager The StateManager
	    '''
    """
    def __init__(self, window):
	self.window = window
	self.stack = []

    def _hide_state(self, state):
	self.window.remove_handlers(state)
	if hasattr(state, 'on_hide_state'):
	    state.on_hide_state(self.window, self)

    def _show_state(self, state):
	self.window.push_handlers(state)
	if hasattr(state, 'on_show_state'):
	    state.on_show_state(self.window, self)

    def push(self, state):
	if self.stack:
	    self._hide_state(self.stack[-1])
	self.stack.append(state)
	self._show_state(self.stack[-1])

    def pop(self):
	assert self.stack
	self._hide_state(self.stack[-1])
	retval = self.stack.pop()
	if self.stack:
	    self._show_state(self.stack[-1])
	else:
	    # no more states to show - we're done!
	    pyglet.app.exit()
	return retval




# kytten/override.py
# Copyrighted (C) 2009 by Conrad "Lynx" Wong

import pyglet

KYTTEN_LAYOUT_GROUPS = {}
KYTTEN_LAYOUT_GROUP_REFCOUNTS = {}

def GetKyttenLayoutGroups(group):
    if not KYTTEN_LAYOUT_GROUPS.has_key(group):
        top_group = pyglet.text.layout.TextLayoutGroup(group)
        background_group = pyglet.graphics.OrderedGroup(0, top_group)
        foreground_group = \
            pyglet.text.layout.TextLayoutForegroundGroup(1, top_group)
        foreground_decoration_group = \
            pyglet.text.layout.TextLayoutForegroundDecorationGroup(
                2, top_group)
        KYTTEN_LAYOUT_GROUPS[group] = (top_group,
                                       background_group,
                                       foreground_group,
                                       foreground_decoration_group)
        KYTTEN_LAYOUT_GROUP_REFCOUNTS[group] = 0
    KYTTEN_LAYOUT_GROUP_REFCOUNTS[group] += 1
    return KYTTEN_LAYOUT_GROUPS[group]

def ReleaseKyttenLayoutGroups(group):
    KYTTEN_LAYOUT_GROUP_REFCOUNTS[group] -= 1
    if not KYTTEN_LAYOUT_GROUP_REFCOUNTS[group]:
        del KYTTEN_LAYOUT_GROUP_REFCOUNTS[group]
        del KYTTEN_LAYOUT_GROUPS[group]

class KyttenLabel(pyglet.text.Label):
    def _init_groups(self, group):
        if not group:
            return # use the default groups
        self.top_group, self.background_group, self.foreground_group, \
            self.foreground_decoration_group = GetKyttenLayoutGroups(group)

    def teardown(self):
        pyglet.text.Label.teardown(self)
        group = self.top_group.parent
        if group is not None:
            ReleaseKyttenLayoutGroups(group)
            self.top_group = self.background_self = self.foreground_group \
                = self.foreground_decoration_group = None

class KyttenInputLabel(KyttenLabel):
    def _get_left(self):
        if self._multiline:
            width = self._width
        else:
            width = self.content_width
            if self.width and width > self.width:
                # align to right edge, clip left
                return self._x + self.width - width

        if self._anchor_x == 'left':
            return self._x
        elif self._anchor_x == 'center':
            return self._x - width // 2
        elif self._anchor_x == 'right':
            return self._x - width
        else:
            assert False, 'Invalid anchor_x'

    def _update(self):
        pyglet.text.Label._update(self)

        # Iterate through our vertex lists and break if we need to clip
        remove = []
        if self.width and not self._multiline:
            for vlist in self._vertex_lists:
                num_quads = len(vlist.vertices) / 8
                remove_quads = 0
                has_quads = False
                for n in xrange(0, num_quads):
                    x1, y1, x2, y2, x3, y3, x4, y4 = vlist.vertices[n*8:n*8+8]
                    tx1, ty1, tz1, tx2, ty2, tz2, \
                       tx3, ty3, tz3, tx4, ty4, tz4 = \
                       vlist.tex_coords[n*12:n*12+12]
                    if x2 >= self._x:
                        has_quads = True
                        m = n - remove_quads  # shift quads left
                        if x1 < self._x:  # clip on left side
                            percent = (float(self._x) - float(x1)) / \
                                      (float(x2) - float(x1))
                            x1 = x4 = max(self._x, x1)
                            tx1 = tx4 = (tx2 - tx1) * percent + tx1
                        vlist.vertices[m*8:m*8+8] = \
                            [x1, y1, x2, y2, x3, y3, x4, y4]
                        vlist.tex_coords[m*12:m*12+12] = \
                             [tx1, ty1, tz1, tx2, ty2, tz2,
                              tx3, ty3, tz3, tx4, ty4, tz4]
                    else:
                        # We'll delete quads entirely not visible
                        remove_quads = remove_quads + 1
                if remove_quads == num_quads:
                    remove.append(vlist)
                elif remove_quads > 0:
                    vlist.resize((num_quads - remove_quads) * 4)
        for vlist in remove:
            vlist.delete()
            self._vertex_lists.remove(vlist)




# kytten/document.py
# Copyrighted (C) 2009 by Conrad "Lynx" Wong

import pyglet

from widgets import Control
from scrollbar import VScrollbar

class Document(Control):
    """
    Allows you to embed a document within the GUI, which includes a
    vertical scrollbar as needed.
    """
    def __init__(self, document, width=1000, height=5000,
                 is_fixed_size=False, always_show_scrollbar=False):
        """
        Creates a new Document.
        """
        Control.__init__(self, width, height)
        self.max_height = height
        self.content_width = width
        if isinstance(document, basestring):
            self.document = pyglet.text.document.UnformattedDocument(document)
        else:
            self.document = document
        self.content = None
        self.content_width = width
        self.scrollbar = None
        self.set_document_style = False
        self.is_fixed_size = is_fixed_size
        self.always_show_scrollbar = always_show_scrollbar
        self.needs_layout = False

    def _do_set_document_style(self, attr, value):
        length = len(self.document.text)
        runs = [(start, end, doc_value) for start, end, doc_value in
                self.document.get_style_runs(attr).ranges(0, length)
                if doc_value is not None]
        if not runs:
            terminator = len(self.document.text)
        else:
            terminator = runs[0][0]
        self.document.set_style(0, terminator, {attr: value})

    def _get_controls(self):
        controls = []
        if self.scrollbar:
            controls += self.scrollbar._get_controls()
        controls += Control._get_controls(self)
        return controls

    def delete(self):
        if self.content is not None:
            self.content.delete()
            self.content = None
        if self.scrollbar is not None:
            self.scrollbar.delete()
            self.scrollbar = None

    def do_set_document_style(self, dialog):
        self.set_document_style = True

        # Check the style runs to make sure we don't stamp on anything
        # set by the user
        self._do_set_document_style('color', dialog.theme['text_color'])
        self._do_set_document_style('font_name', dialog.theme['font'])
        self._do_set_document_style('font_size', dialog.theme['font_size'])

    def get_text(self):
        return self.document.text

    def layout(self, x, y):
        self.x, self.y = x, y
        self.content.begin_update()
        self.content.x = x
        self.content.y = y
        self.content.end_update()
        if self.scrollbar is not None:
            self.scrollbar.layout(x + self.content_width, y)

    def on_update(self, dt):
        """
        On updates, we update the scrollbar and then set our view offset
        if it has changed.

        @param dt Time passed since last update event (in seconds)
        """
        if self.scrollbar is not None:
            self.scrollbar.dispatch_event('on_update', dt)
            pos = self.scrollbar.get(self.max_height,
                                     self.content.content_height)
            if pos != -self.content.view_y:
                self.content.view_y = -pos

        if self.needs_layout:
            self.needs_layout = False
            self.saved_dialog.set_needs_layout()

    def size(self, dialog):
        if dialog is None:
            return

        Control.size(self, dialog)
        if not self.set_document_style:
            self.do_set_document_style(dialog)
        if self.content is None:
            self.content = pyglet.text.layout.IncrementalTextLayout(
                self.document,
                self.content_width,
                self.max_height,
                multiline=True, batch=dialog.batch, group=dialog.fg_group)
            if self.is_fixed_size or (self.max_height and
                self.content.content_height > self.max_height):
                self.height = self.max_height
            else:
                self.height = self.content.content_height
            self.content.height = self.height
        if self.always_show_scrollbar or \
           (self.max_height and self.content.content_height > self.max_height):
            if self.scrollbar is None:
                self.scrollbar = VScrollbar(self.max_height)
            self.scrollbar.size(dialog)
            self.scrollbar.set(self.max_height, self.content.content_height)
        if self.scrollbar is not None:
            self.width = self.content_width + self.scrollbar.width
        else:
            self.width = self.content_width

    def set_text(self, text):
        self.document.text = text
        self.needs_layout = True


# theme_editor.py
# Copyrighted (C) 2009 by Conrad "Lynx" Wong

# Creates or modifies a .json file for a kytten GUI theme.

import glob
import os
import sys

import pyglet
# Disable error checking for increased performance
pyglet.options['debug_gl'] = False
from pyglet import gl

try:
    import kytten
except ImportError:
    # Workaround for editor being in sibling directory to kytten
    sys.path.append(os.path.dirname(sys.path[0]))
    import kytten

from kytten import safe_eval
from kytten.scrollable import ScrollableGroup
from kytten.theme import TextureGraphicElementTemplate
from kytten.theme import FrameTextureGraphicElementTemplate
from kytten.widgets import Control
from proxy_dialog import ProxyDialog
from resizable import Resizable
from image_region_placer import ImageRegionPlacer
from state_manager import StateManager
from color_selector import ColorSelector

# Default theme
theme_dir = os.path.join(os.getcwd(), 'theme')
if not os.path.isfile(theme_dir) and not os.path.isdir(theme_dir):
    # Workaround for editor being in sibling directory to theme
    theme_dir = os.path.join(os.path.dirname(sys.path[0]), 'theme')
gTheme = kytten.Theme(theme_dir, override={
    'gui_color': (255, 235, 128, 255),
})

# Global Variables

gDirty = False

#
# Theme Editor States
#

class NoTexturesException(Exception):
    pass

class BaseState:
    def __init__(self):
	self.dialog = None
	self.window = None
	self.manager = None
	self.popup = None

    def on_draw(self):
	if self.window is not None:
	    self.window.clear()
	if self.dialog is not None:
	    self.dialog.draw()
	if self.popup is not None:
	    self.popup.draw()

    def on_hide_state(self, window, manager):
	if self.popup is not None:
	    self.popup.teardown()
	    self.popup = None
	if self.dialog is not None:
	    self.dialog.teardown()
	    self.dialog = None

    def on_show_state(self, window, manager):
	self.window = window
	self.manager = manager

    def popup_message(self, message):
	def delete_popup(dialog):
		    self.popup = None
	if self.popup is not None:
	    self.popup.teardown()
	self.popup = kytten.PopupMessage(
	    message, theme=gTheme, window=window, on_escape=delete_popup)

class ThemeDirSelectState(BaseState):
    def on_show_state(self, window, manager):
	BaseState.on_show_state(self, window, manager)

	def on_theme_dir_select(filename):
	    if filename:
		self.select_theme_dir(filename)
	    else:
		self.popup_message("Please select a directory")

	def on_theme_dir_cancel(dialog):
	    self.manager.pop()

	self.dialog = kytten.DirectorySelectDialog(
	    title="Select a theme directory",
	    window=window,
	    anchor=kytten.ANCHOR_CENTER,
	    theme=gTheme,
	    on_select=on_theme_dir_select,
	    on_escape=on_theme_dir_cancel)

    def select_theme_dir(self, filename):
	json_files = glob.glob(os.path.join(filename, '*.json'))
	if len(json_files) < 2:
	    if json_files:
		json_filename = os.path.basename(json_files[0])  # first file
		try:
		    self.manager.push(ThemeEditState(filename, json_filename))
		except NoTexturesException:
		    self.popup_message("Directory must contain textures")
	    else:
		self.manager.push(ThemeNewFileState(filename))
	else:
	    self.manager.push(ThemeFileSelectState(filename))

class ThemeFileSelectState(BaseState):
    def __init__(self, theme_dir):
	BaseState.__init__(self)
	self.theme_dir = theme_dir

    def on_show_state(self, window, manager):
	BaseState.on_show_state(self, window, manager)

	def on_theme_file_select(filename):
	    if filename:
		self.select_theme_file(filename)
	    else:
		self.popup_message("Please select a file")

	def on_theme_file_cancel(dialog):
	    self.manager.pop()

	json_files = glob.glob(os.path.join(self.theme_dir, '*.json'))
	json_files = [os.path.basename(x) for x in json_files]
	self.dialog = kytten.Dialog(
	    kytten.TitleFrame("kytten Theme Editor",
		kytten.VerticalLayout([
		    kytten.Label("Please select a theme file"),
		    kytten.Menu(options=['(New File)'] + json_files,
				on_select=on_theme_file_select)
		]),
	    ),
	    window=window,
	    anchor=kytten.ANCHOR_CENTER,
	    theme=gTheme,
	    on_escape=on_theme_file_cancel)

    def select_theme_file(self, filename):
	if filename == '(New File)':
	    self.manager.push(ThemeNewFileState(self.theme_dir))
	else:
	    try:
		self.manager.push(ThemeEditState(self.theme_dir, filename))
	    except NoTexturesException:
		self.popup_message('Directory must contain textures')

class ThemeNewFileState(BaseState):
    def __init__(self, theme_dir):
	BaseState.__init__(self)
	self.theme_dir = theme_dir

    def on_show_state(self, window, manager):
	BaseState.on_show_state(self, window, manager)

	def on_theme_file_cancel(dialog):
	    self.manager.pop()

	def on_theme_file_enter(dialog):
	    form = dialog.get_values()
	    filename = form['name']
	    if filename:
		name = os.path.join(self.theme_dir, filename)
		if os.path.isfile(name):
		    self.popup_message("%s already exists" % filename)
		elif not filename.endswith('.json'):
		    self.popup_message("%s must end in .json" % filename)
		else:
		    self.set_theme_file(filename)
	    else:
		self.popup_message("Please select a file")

	self.dialog = kytten.Dialog(
	    kytten.TitleFrame("kytten Theme Editor",
		kytten.VerticalLayout([
		    kytten.Label("Please select a theme file name"),
		    kytten.Input(id='name')
		]),
	    ),
	    window=window,
	    anchor=kytten.ANCHOR_CENTER,
	    theme=gTheme,
	    on_escape=on_theme_file_cancel,
	    on_enter=on_theme_file_enter)

    def set_theme_file(self, filename):
	try:
	    new_state = ThemeEditState(self.theme_dir, filename)
	    self.manager.pop()
	    self.manager.push(new_state)
	except NoTexturesException:
	    self.popup_message("Directory must contain textures")

class _NodeEditState(BaseState):
    def __init__(self, theme_dir, theme, textures, path=[],
		 on_exit=None):
	BaseState.__init__(self)
	self.theme_dir = theme_dir
	self.theme = theme
	self.textures = textures
	self.path = path
	self.add_button = None
	self.delete_button = None
	self.on_exit = on_exit
	self.file_label = None

    def _get_buttons(self):
	if self.path:
	    disabled = bool(self.theme.get(self.path))
	    def on_return():
		self.manager.pop()
	    def on_delete_click():
		self.do_delete_this()
	    self.delete_button = kytten.Button("Delete Component",
					       disabled=disabled,
					       on_click=on_delete_click)
	    return [kytten.Button("Back", on_click=on_return),
		    self.delete_button]
	else:
	    return []

    def _get_content(self):
	content = [kytten.Label("Theme: %s" % self.theme_dir)]
	if self.path:
	    content += [kytten.Label("Path: %s" % '/'.join(self.path))]
	else:
	    assert hasattr(self, 'name')
	    self.file_label = kytten.Label("File: %s" % self.name)
	    content += [self.file_label]
	content += [kytten.HorizontalLayout(self._get_buttons())]
	content += self._get_image_fields()
	content += self._get_custom_components()
	content += self._get_custom_fields()
	return content

    def _get_custom_components(self):
	# Handle all components without built-in handlers here
	components = []
	items = self.theme[self.path].items()
	items.sort(lambda x, y: cmp(x[0], y[0]))
	for k, v in items:
	    if isinstance(v, dict):
		components.append(k)
	def edit_component(choice):
	    self.do_edit_component(choice)
	def on_add_click():
	    self.do_add_new_component()
	return [kytten.FoldingSection("Custom Components",
	    kytten.VerticalLayout([
		kytten.Menu(options=components,
			    on_select=edit_component,
			    align=kytten.HALIGN_LEFT),
	        kytten.Button("Add Component",
			      on_click=on_add_click)]))]

    def _get_custom_fields(self):
	# Handle all fields here that we don't have a basic handler for
	fields_layout = None
	def on_input(id, value):
	    global gDirty
	    gDirty = True
	    try:
		self.theme[self.path][id] = safe_eval.safe_eval(value)
	    except kytten.safe_eval.Unsafe_Source_Error:
		self.theme[self.path][id] = value
	def on_delete(id):
	    self.do_delete_field(fields_layout, id)
	fields = []
	items = self.theme[self.path].items()
	items.sort(lambda x, y: cmp(x[0], y[0]))
	index = 0
	for k, v in items:
	    if not k.startswith('image') and not isinstance(v, dict):
		if not k.endswith('color'):
		    fields.append([kytten.Label(k),
				   kytten.Input(id=k, text=str(v),
						on_input=on_input),
				   kytten.Button("Delete", id=k,
						 on_click=on_delete)])
		else:
		    fields.append([kytten.Label(k),
				   ColorSelector(id=k, color=v,
						 on_select=on_input),
				   kytten.Button("Delete", id=k,
						 on_click=on_delete)])
	    index += 1
	fields_layout = kytten.GridLayout(fields)

	# Pop up a dialog if we want to add a new field
	def on_add_field():
	    self.do_add_new_field(fields_layout, on_input, on_delete)
	return [kytten.FoldingSection("Custom Fields",
		    kytten.VerticalLayout([
			kytten.GridLayout(fields, anchor=kytten.ANCHOR_LEFT),
			kytten.Button("Add Field", on_click=on_add_field),
		    ]))]

    def _get_image_fields(self):
	# Display a menu to choose images to edit
	images  = []
	items = self.theme[self.path].items()
	items.sort(lambda x, y: cmp(x[0], y[0]))
	for k, v in items:
	    if isinstance(v, TextureGraphicElementTemplate):
		assert k.startswith('image')
		images.append(k)
	def edit_image(choice):
	    self.do_edit_image(choice)
	def add_image():
	    self.do_add_new_image()
	return [kytten.FoldingSection("Images",
	    kytten.VerticalLayout([
		kytten.Menu(options=images,
			    on_select=edit_image,
			    align=kytten.HALIGN_LEFT),
	        kytten.Button('Add Image', on_click=add_image)]))]

    def do_add_new_component(self):
	if self.popup is not None:
	    self.popup.teardown()
	def do_cancel_add(dialog=None):
	    self.popup.teardown()
	    self.popup = None
	def do_add_component(dialog=None):
	    global gDirty
	    gDirty = True

	    form = self.popup.get_values()
	    self.popup.teardown()
	    self.popup = None

	    name = form['name']
	    if self.theme[self.path].has_key(name):
		self.popup_message("%s is already a field name!" % name)
		return
	    self.theme[self.path][name] = {}
	    self.manager.push(_NodeEditState(self.theme_dir,
					     self.theme,
					     self.textures,
					     self.path + [name]))

	self.popup = kytten.Dialog(
	    kytten.Frame(
		kytten.VerticalLayout([
		    kytten.HorizontalLayout([
			kytten.Label('Name'),
			kytten.Input(id='name', text='new_component')]),
		    kytten.HorizontalLayout([
			kytten.Button("Add", on_click=do_add_component),
			None,
			kytten.Button("Cancel", on_click=do_cancel_add)
		    ])])),
	    window=self.window, theme=gTheme,
	    on_enter=do_add_component, on_escape=do_cancel_add)

    def do_add_new_field(self, fields_layout, on_input, on_delete):
	if self.popup is not None:
	    self.popup.teardown()
	def do_cancel_add(dialog=None):
	    self.popup.teardown()
	    self.popup = None
	def do_add_field(dialog=None):
	    global gDirty
	    gDirty = True

	    form = self.popup.get_values()
	    self.popup.teardown()
	    self.popup = None

	    name = form['name']
	    if self.theme[self.path].has_key(name):
		self.popup_message("%s is already a field name!" % name)
		return
	    try:
		value = safe_eval.safe_eval(form['value'])
	    except kytten.safe_eval.Unsafe_Source_Error:
		value = form['value']
	    self.theme[self.path][name] = value
	    fields_layout.add_row(
		[kytten.Label(name),
		 kytten.Input(id=name, text=str(value), on_input=on_input),
		 kytten.Button("Delete", id=name, on_click=on_delete)])

	    if not self.delete_button.is_disabled():
		self.delete_button.disable()
	    self.dialog.set_needs_layout()

	self.popup = kytten.Dialog(
	    kytten.Frame(
		kytten.VerticalLayout([
		    kytten.GridLayout([
			[kytten.Label('Name'),
			 kytten.Input(id='name', text='new_field')],
			[kytten.Label('Value'),
			 kytten.Input(id='value', text='new_value')]]),
		    kytten.HorizontalLayout([
			kytten.Button("Add", on_click=do_add_field),
			None,
			kytten.Button("Cancel", on_click=do_cancel_add)
		    ])])),
	    window=self.window, theme=gTheme,
	    on_enter=do_add_field, on_escape=do_cancel_add)

    def do_add_new_image(self):
	if self.popup is not None:
	    self.popup.teardown()
	def do_cancel_add(dialog=None):
	    self.popup.teardown()
	    self.popup = None
	def do_add_image(dialog=None):
	    global gDirty
	    gDirty = True

	    form = self.popup.get_values()
	    self.popup.teardown()
	    self.popup = None

	    name = form['name']
	    if self.theme[self.path].has_key(name):
		self.popup_message("%s is already a field name!" % name)
		return
	    if not name.startswith('image'):
		self.popup_message("Image names must begin with 'image'")
	    texture = self.textures[form['texture']]
	    self.theme[self.path][name] = FrameTextureGraphicElementTemplate(
		self.theme,
		texture,
		[0, 0, texture.width, texture.height],  # stretch
		[0, 0, 0, 0])                           # padding
	    self.manager.push(ImageEditState(self.theme_dir,
					     self.theme,
					     self.textures,
					     self.path,
					     name))

	self.popup = kytten.Dialog(
	    kytten.Frame(
		kytten.VerticalLayout([
		    kytten.GridLayout([
			[kytten.Label('Name'),
			 kytten.Input(id='name', text='image')],
			[kytten.Label('Texture'),
			 kytten.Dropdown(id='texture',
					 options=self.textures.keys(),
					 selected=self.textures.keys()[0])]
		    ]),
		    kytten.HorizontalLayout([
			kytten.Button("Add", on_click=do_add_image),
			None,
			kytten.Button("Cancel", on_click=do_cancel_add)
		    ])])),
	    window=self.window, theme=gTheme,
	    on_enter=do_add_image, on_escape=do_cancel_add)

    def do_delete_field(self, fields_layout, id):
	global gDirty
	gDirty = True

	del self.theme[self.path][id]
	index = 0
	for row in fields_layout.content:
	    if row[1].id == id:
		fields_layout.delete_row(index)
		break
	    index += 1
	if not self.theme[self.path]:
	    if self.delete_button.is_disabled():
		self.delete_button.enable()

    def do_delete_this(self):
	if self.theme[self.path]:
	    self.popup_message("Component must be empty to be deleted")
	else:
	    global gDirty
	    gDirty = True

	    del self.theme[self.path[:-1]][self.path[-1]]
	    self.manager.pop()

    def do_edit_component(self, component):
	self.manager.push(_NodeEditState(self.theme_dir,
					 self.theme,
					 self.textures,
					 self.path + [component]))

    def do_edit_image(self, image):
	self.manager.push(ImageEditState(self.theme_dir, self.theme,
					 self.textures, self.path, image))

    def on_show_state(self, window, manager):
	BaseState.on_show_state(self, window, manager)

	content = self._get_content()
	def on_escape(dialog):
	    self.manager.pop()
	self.dialog = kytten.Dialog(
	    kytten.TitleFrame('kytten Theme Editor',
		kytten.Scrollable(
		    kytten.VerticalLayout(content, align=kytten.HALIGN_LEFT),
		    height=500)
	    ),
	    window=window,
	    anchor=kytten.ANCHOR_CENTER,
	    theme=gTheme,
	    on_escape=self.on_exit or on_escape)

class ThemeEditState(_NodeEditState):
    def __init__(self, theme_dir, name='theme.json'):
	self.name = name
	theme = kytten.Theme(theme_dir, allow_empty_theme=True, name=name)
	textures = {}
	files = glob.glob(os.path.join(theme_dir, '*.png')) + \
		glob.glob(os.path.join(theme_dir, '*.gif'))
	for f in [os.path.basename(x) for x in files]:
	    textures[f] = theme._get_texture(f)
	if not textures:
	    raise NoTexturesException

	_NodeEditState.__init__(self, theme_dir, theme, textures,
				on_exit=self.do_exit)

    def _get_buttons(self):
	def on_save():
	    self.do_save()
	def on_save_as():
	    self.do_save_as()
	def on_new():
	    if gDirty:
		if self.popup is not None:
		    self.popup.teardown()
		self.popup = kytten.PopupConfirm(
		    "Create new theme without saving?",
		    on_ok=self.do_new,
		    theme=gTheme,
		    window=self.window)
	    else:
		self.do_new()
	return [kytten.Button("Save", on_click=on_save),
		kytten.Button("Save As", on_click=on_save_as),
		kytten.Button("New", on_click=on_new)]

    def do_exit(self, dialog=None):
	if gDirty:
	    if self.popup is not None:
		self.popup.teardown()
	    self.popup = kytten.PopupConfirm("Exit without saving?",
					     on_ok=self.do_really_exit,
					     theme=gTheme,
					     window=self.window)
	else:
	    self.do_really_exit()

    def do_new(self, dialog=None):
	self.manager.pop()
	self.manager.push(ThemeNewFileState(self.theme_dir))

    def do_really_exit(self, dialog=None):
	self.manager.pop()

    def do_save(self):
	global gDirty
	gDirty = False

	filename = os.path.join(self.theme_dir, self.name)
	new_filename = os.path.join(self.theme_dir, '%s-new' % self.name)
	old_filename = os.path.join(self.theme_dir, '%s-old' % self.name)
	if os.path.isfile(old_filename):
	    os.unlink(old_filename)
	f = open(new_filename, 'w')
	self.theme.write(f)
	f.close()
	if os.path.isfile(filename):
	    os.rename(filename, old_filename)
	os.rename(new_filename, filename)
	if os.path.isfile(old_filename):
	    os.unlink(old_filename)

	self.popup_message("Saved %s" % self.name)

    def do_save_as(self):
	if self.popup is not None:
	    self.popup.teardown()
	def do_cancel_save_as(dialog):
	    self.popup.teardown()
	    self.popup = None
	def do_set_filename(dialog):
	    form = dialog.get_values()
	    filename = form['name']
	    if filename:
		name = os.path.join(self.theme_dir, filename)
		if os.path.isfile(name):
		    self.popup_message("%s already exists" % filename)
		elif not filename.endswith('.json'):
		    self.popup_message("%s must end in .json" % filename)
		else:
		    self.name = filename
		    self.file_label.label.text = 'File: %s' % self.name
		    self.do_save()
	    else:
		self.popup_message("Please select a file")
	self.popup = kytten.Dialog(
	    kytten.Frame(
		kytten.HorizontalLayout([
		    kytten.Label("Filename"),
		    kytten.Input(id='name')
		])),
	    theme=gTheme, window=self.window,
	    on_enter=do_set_filename,
	    on_escape=do_cancel_save_as)

class ImageEditState(BaseState):
    def __init__(self, theme_dir, theme, textures, path, image):
	BaseState.__init__(self)
	self.theme_dir = theme_dir
	self.theme = theme
	self.textures = textures
	self.path = path
	self.image = image
	self.template = None
	self.region = [0, 0, 0, 0]
	self.stretch = [0, 0, 0, 0]
	self.padding = [0, 0, 0, 0]
	self.state = 'Region'

    def _get_content(self):
	def on_return():
	    self.manager.pop()
	def on_delete():
	    self.do_delete_image()
	def on_change_texture():
	    self.do_change_texture()
	content = [kytten.VerticalLayout([
	    kytten.Document("""Theme: %s
Path: %s
Image: %s
""" % (self.theme_dir, '/'.join(self.path), self.image), width=700),
	    kytten.HorizontalLayout([
		kytten.Button("Back", on_click=on_return),
		kytten.Button("Change Texture", on_click=on_change_texture),
		kytten.Button("Delete Image", on_click=on_delete)
	    ]),
	], align=kytten.HALIGN_LEFT)]

	# Locate the base texture for the template
	self.template = self.theme[self.path][self.image]
	texture_id = self.template.texture.id
	our_texture = None
	our_filename = None
	for texture_name, texture in self.textures.iteritems():
	    if texture.id == texture_id:
		our_filename = texture_name
		our_texture = texture
		break
	assert our_texture is not None

	# Determine the region that we occupy
	x, y = self.template.texture.x, self.template.texture.y
	width, height = self.template.width,self. template.height
	self.region = (x, y, width, height)
	left, right, top, bottom = self.template.margins
	self.stretch = (left, bottom,
			width - right - left, height - top - bottom)
	self.padding = self.template.padding

	# Create the Resizable for the example's content
	example_resizable = Resizable(100, 100)
	def enable_example_resizable(is_checked):
	    if is_checked:
		example_resizable.enable()
	    else:
		example_resizable.disable()

	# Create the example
	example = kytten.Frame(example_resizable,
			       path=self.path,
			       image_name=self.image)

	# Create the ImageRegionPlacer for the editor
	region_placer = None
	def set_region(x, y, width, height):
	    global gDirty
	    gDirty = True

	    if self.state == 'Region':
		# Try to keep stretch area the same but restrict it to
		# the current available region
		ox, oy, _, _ = self.region
		sx, sy, swidth, sheight = self.stretch
		sx += ox
		sy += oy
		if sx < x:
		    swidth = max(0, swidth - x + sx)
		    sx = x
		elif sx > x + width:
		    swidth = 0
		    sx = x + width
		if sy < y:
		    sheight = max(0, sheight - y + sy)
		    sy = y
		elif sy > y + height:
		    sheight = 0
		    sy = y + height
		if (sx - x) + swidth > width:
		    swidth = max(0, width - (sx - x))
		if (sy - y) + sheight > height:
		    sheight = max(0, height - (sy - y))

		self.region = (x, y, width, height)
		self.stretch = (sx - x, sy - y, swidth, sheight)
	    elif self.state == 'Stretch':
		rx, ry, rwidth, rheight = self.region
		self.stretch = (x - rx, y - ry, width, height)
	    else:  # Padding
		rx, ry, rwidth, rheight = self.region
		left = x - rx
		bottom = y - ry
		top = ry + rheight - height - y
		right = rx + rwidth - width - x
		self.padding = (left, right, top, bottom)

	    # Drop our old template and construct a new one
	    x, y, width, height = self.region
	    texture = self.theme._get_texture_region(our_filename,
						     x, y, width, height)
	    self.template = FrameTextureGraphicElementTemplate(
		self.theme, texture, self.stretch, self.padding)
	    self.theme[self.path][self.image] = self.template
	    example.delete()
	    self.dialog.set_needs_layout()
	region_placer = ImageRegionPlacer(texture, x, y, width, height,
					  on_resize=set_region)
	def set_placer_scale(scale):
	    region_placer.set_scale(scale)

	# Create a drop-down to control what region we're setting
	def on_region_select(choice):
	    self.state = choice
	    if choice == 'Region':
		region_placer.set_region(
		    *self.region,
		    color=ImageRegionPlacer.IMAGE_REGION_COLOR)
	    elif choice == 'Stretch':
		rx, ry, rwidth, rheight = self.region
		x, y, width, height = self.stretch
		region_placer.set_region(
		    x + rx, y + ry, width, height,
		    color=ImageRegionPlacer.IMAGE_STRETCH_COLOR,
		    limits=self.region)
	    else:  # Padding
		rx, ry, rwidth, rheight = self.region
		left, right, top, bottom = self.padding
		region_placer.set_region(
		    left + rx, bottom + ry,
		    rwidth - left - right,
		    rheight - top - bottom,
		    color=ImageRegionPlacer.IMAGE_PADDING_COLOR,
		    limits=self.region)
	region_control = kytten.Dropdown(
	    options=['Region', 'Stretch', 'Padding'],
	    selected=self.state,
	    on_select=on_region_select)

	content += [
	    kytten.FoldingSection('Image Region',
		kytten.VerticalLayout([
		    region_control,
		    region_placer,
		    kytten.GridLayout([
			[kytten.Label('Scale'),
			 kytten.Slider(value=1.0,
				       min_value=1.0, max_value=8.0,
				       steps=7, width=400,
				       on_set=set_placer_scale)]
		    ], anchor=kytten.ANCHOR_LEFT)
		], padding=10)),
	    kytten.FoldingSection('Example',
		kytten.VerticalLayout([
		    ProxyDialog(example,
				self.theme),
		    kytten.Checkbox("Show content sizer",
				    on_click=enable_example_resizable,
				    is_checked=True),
		])),
	]
	return content

    def do_change_texture(self):
	if self.popup is not None:
	    self.popup.teardown()
	def do_cancel_change(dialog=None):
	    self.popup.teardown()
	    self.popup = None
	def do_change_texture(dialog=None):
	    global gDirty
	    gDirty = True

	    form = self.popup.get_values()
	    self.popup.teardown()
	    self.popup = None

	    texture = self.textures[form['texture']]
	    self.theme[self.path][self.image] = \
		FrameTextureGraphicElementTemplate(
		    self.theme,
		    texture,
		    [0, 0, texture.width, texture.height],  # stretch
		    [0, 0, 0, 0])                           # padding
	    manager = self.manager
	    manager.pop()
	    manager.push(ImageEditState(self.theme_dir,
					     self.theme,
					     self.textures,
					     self.path,
					     self.image))

	self.popup = kytten.Dialog(
	    kytten.Frame(
		kytten.VerticalLayout([
		    kytten.HorizontalLayout([
			kytten.Label('Texture'),
			kytten.Dropdown(id='texture',
					options=self.textures.keys(),
					selected=self.textures.keys()[0])]),
		    kytten.HorizontalLayout([
			kytten.Button("Change", on_click=do_change_texture),
			None,
			kytten.Button("Cancel", on_click=do_cancel_change)
		    ]),
		])),
	    window=self.window, theme=gTheme,
	    on_enter=do_change_texture, on_escape=do_cancel_change)

    def do_delete_image(self):
	global gDirty
	gDirty = True

	del self.theme[self.path][self.image]
	self.manager.pop()

    def on_show_state(self, window, manager):
	BaseState.on_show_state(self, window, manager)

	def on_escape(dialog):
	    self.manager.pop()
	content = self._get_content()
	self.dialog = kytten.Dialog(
	    kytten.TitleFrame("kytten Image Editor",
		kytten.Scrollable(
		    kytten.VerticalLayout(content),
		    width=750, height=500)
	    ),
	    window=window,
	    anchor=kytten.ANCHOR_CENTER,
	    theme=gTheme,
	    on_escape=on_escape)

if __name__ == '__main__':
    window = pyglet.window.Window(
	800, 600, caption='Kytten Theme Editor', resizable=True, vsync=True)
    batch = pyglet.graphics.Batch()
    bg_group = pyglet.graphics.OrderedGroup(0)
    fg_group = pyglet.graphics.OrderedGroup(1)

    # Update as often as possible (limited by vsync, if not disabled)
    window.register_event_type('on_update')
    def update(dt):
	window.dispatch_event('on_update', dt)
    pyglet.clock.schedule(update)

    # StateManager keeps track of what we're doing
    manager = StateManager(window)

    # Start off by picking the theme directory
    manager.push(ThemeDirSelectState())

    pyglet.app.run()



# image_region_placer.py
# Copyrighted (C) 2009 by Conrad "Lynx" Wong

import pyglet
from pyglet import gl

import kytten
from kytten.widgets import Control

class ImageRegionPlacer(Control):
    """
    Display an image and a region within the image, which can be
    resized or moved.
    """
    IMAGE_REGION_COLOR = (255, 128, 64, 255)
    IMAGE_STRETCH_COLOR = (64, 255, 128, 255)
    IMAGE_PADDING_COLOR = (64, 128, 255, 255)

    def __init__(self, texture, x=0, y=0, width=0, height=0, scale=1,
		 on_resize=None):
	Control.__init__(self)
	self.texture = texture
	self.on_resize = on_resize
	self.texture_group = None
	self.texture_vlist = None
	self.resizer_vlist = None
	self.limits_vlist = None
	self.region = [x, y, width, height]
	self.limits = None
	self.scale = scale
	self.color = self.IMAGE_REGION_COLOR
	self.is_dragging = False
	self.corner = None

    def _get_controls(self):
	_, _, width, height = self.region
	if width < 20 or height < 20:
	    return [(self, self.x - 8, self.x + self.width + 8,
		           self.y + self.height + 8, self.y - 8)]
	else:
	    return Control._get_controls(self)

    def _get_limits_vertices(self):
	if self.limits is not None:
	    x, y, width, height = self.limits
	    x0 = int(self.x + x * self.scale)
	    y0 = int(self.y + y * self.scale)
	    x1 = x0 + int(width * self.scale)
	    y1 = y0 + int(height * self.scale)
	else:
	    x0, y0 = int(self.x), int(self.y)
	    x1 = x0 + int(self.width)
	    y1 = y0 + int(self.height)
	return [x0, y0, x1, y0, x1, y0, x1, y1,
		x1, y1, x0, y1, x0, y1, x0, y0]

    def _get_resizer_vertices(self):
	x, y, width, height = self.region
	if width < 20 or height < 20:  # sizer corners on outside
	    x0 = int(self.x + x * self.scale - 8)
	    y0 = int(self.y + y * self.scale - 8)
	    x1, y1 = x0 + 8, y0 + 8
	    x2 = x1 + int(width * self.scale)
	    y2 = y1 + int(height * self.scale)
	    x3, y3 = x2 + 8, y2 + 8
	    return [x1, y1, x2, y1, x2, y1, x2, y2, # center square
		    x2, y2, x1, y2, x1, y2, x1, y1,
		    x0, y0, x1, y0, x1, y0, x1, y1, # lower left
		    x1, y1, x0, y1, x0, y1, x0, y0,
		    x2, y0, x3, y0, x3, y0, x3, y1, # lower right
		    x3, y1, x2, y1, x2, y1, x2, y0,
		    x2, y2, x3, y2, x3, y2, x3, y3, # upper right
		    x3, y3, x2, y3, x2, y3, x2, y2,
		    x0, y2, x1, y2, x1, y2, x1, y3, # upper left
		    x1, y3, x0, y3, x0, y3, x0, y2]
	else:                          # sizer corners on inside
	    x0 = int(self.x + x * self.scale)
	    y0 = int(self.y + y * self.scale)
	    x1, y1 = x0 + 8, y0 + 8
	    x3 = x0 + int(width * self.scale)
	    y3 = y0 + int(height * self.scale)
	    x2, y2 = x3 - 8, y3 - 8
	    return [x0, y0, x3, y0, x3, y0, x3, y3, # center square
		    x3, y3, x0, y3, x0, y3, x0, y0,
		    x0, y0, x1, y0, x1, y0, x1, y1, # lower left
		    x1, y1, x0, y1, x0, y1, x0, y0,
		    x2, y0, x3, y0, x3, y0, x3, y1, # lower right
		    x3, y1, x2, y1, x2, y1, x2, y0,
		    x2, y2, x3, y2, x3, y2, x3, y3, # upper right
		    x3, y3, x2, y3, x2, y3, x2, y2,
		    x0, y2, x1, y2, x1, y2, x1, y3, # upper left
		    x1, y3, x0, y3, x0, y3, x0, y2]

    def _get_texture_vertices(self):
	x1, y1 = int(self.x), int(self.y)
	x2 = x1 + int(self.width)
	y2 = y1 + int(self.height)
	return [x1, y1, x2, y1, x2, y2, x1, y2]

    def delete(self):
	Control.delete(self)
	if self.texture_vlist is not None:
	    self.texture_vlist.delete()
	    self.texture_vlist = None
	if self.resizer_vlist is not None:
	    self.resizer_vlist.delete()
	    self.resizer_vlist = None
	if self.limits_vlist is not None:
	    self.limits_vlist.delete()
	    self.limits_vlist = None
	self.texture_group = None

    def layout(self, x, y):
	Control.layout(self, x, y)
	self.texture_vlist.vertices = self._get_texture_vertices()
	self.resizer_vlist.vertices = self._get_resizer_vertices()
	self.limits_vlist.vertices = self._get_limits_vertices()

    def on_mouse_drag(self, x, y, dx, dy, buttons, modifiers):
	if not self.is_dragging:
	    return

	dx = float(dx) / self.scale
	dy = float(dy) / self.scale
	x, y, width, height = self.region
	left = x
	right = x + width
	bottom = y
	top = y + height
	if self.corner == kytten.ANCHOR_TOP_LEFT:  # stretch up and left
	    x = min(x + dx, right)
	    width = right - x
	    height = max(height + dy, 0)
	elif self.corner == kytten.ANCHOR_LEFT:  # stretch left only
	    x = min(x + dx, x + width)
	    width = right - x
	elif self.corner == kytten.ANCHOR_BOTTOM_LEFT:  # down and left
	    x = min(x + dx, right)
	    y = min(y + dy, top)
	    width = right - x
	    height = top - y
	elif self.corner == kytten.ANCHOR_TOP:  # stretch up only
	    height = max(height + dy, 0)
	elif self.corner == kytten.ANCHOR_BOTTOM:  # stretch down only
	    y = min(y + dy, top)
	    height = top - y
	elif self.corner == kytten.ANCHOR_TOP_RIGHT:  # stretch up and right
	    width = max(width + dx, 0)
	    height = max(height + dy, 0)
	elif self.corner == kytten.ANCHOR_RIGHT:  # stretch right only
	    width = max(width + dx, 0)
	elif self.corner == kytten.ANCHOR_BOTTOM_RIGHT:  # down and right
	    y = min(y + dy, top)
	    width = max(width + dx, 0)
	    height = top - y
	else:  # center or no corner set, drag the whole square!
	    x += dx
	    y += dy
	self.set_region(x, y, width, height, self.color, self.limits)

    def on_mouse_press(self, x, y, button, modifiers):
	self.is_dragging = True
	rx, ry, width, height = self.region
	if width < 20 or height < 20:  # sizer corners on outside
	    x0 = int(self.x + rx * self.scale - 8)
	    y0 = int(self.y + ry * self.scale - 8)
	    x1, y1 = x0 + 8, y0 + 8
	    x2 = x1 + int(width * self.scale)
	    y2 = y1 + int(height * self.scale)
	    x3, y3 = x2 + 8, y2 + 8
	else:                          # sizer corners on inside
	    x0 = int(self.x + rx * self.scale)
	    y0 = int(self.y + ry * self.scale)
	    x1, y1 = x0 + 8, y0 + 8
	    x3 = x0 + int(width * self.scale)
	    y3 = y0 + int(height * self.scale)
	    x2, y2 = x3 - 8, y3 - 8
	self.corner = None
	if x < x1:
	    if y < y1: # lower left corner
		self.corner = kytten.ANCHOR_BOTTOM_LEFT
	    elif y > y2: # upper left corner
		self.corner = kytten.ANCHOR_TOP_LEFT
	    else:
		self.corner = kytten.ANCHOR_LEFT
	elif x > x2:
	    if y < y1: # lower right corner
		self.corner = kytten.ANCHOR_BOTTOM_RIGHT
	    elif y > y2: # upper right corner
		self.corner = kytten.ANCHOR_TOP_RIGHT
	    else:
		self.corner = kytten.ANCHOR_RIGHT
	else:
	    if y < y1: # bottom side
		self.corner = kytten.ANCHOR_BOTTOM
	    elif y > y2: # top side
		self.corner = kytten.ANCHOR_TOP

    def on_mouse_release(self, x, y, button, modifiers):
	self.is_dragging = False
	x, y, width, height = self.region
	self.region = [int(x + 0.5), int(y + 0.5),
		       int(width + 0.5), int(height + 0.5)]
	if self.resizer_vlist is not None:
	    self.resizer_vlist.vertices = self._get_resizer_vertices()
	if self.on_resize is not None:
	    self.on_resize(*self.region)

    def set_region(self, x, y, width, height, color, limits=None):
	self.limits = limits
	self.color = color
	if self.limits is None:
	    left_limit = bottom_limit = 0
	    right_limit, top_limit = self.texture.width, self.texture.height
	else:
	    lx, ly, lwidth, lheight = self.limits
	    left_limit = lx
	    bottom_limit = ly
	    right_limit = left_limit + lwidth
	    top_limit = bottom_limit + lheight
	x = min(max(x, left_limit), right_limit)
	y = min(max(y, bottom_limit), top_limit)
	width = min(width, right_limit - x)
	height = min(height, top_limit - y)
	self.region = [x, y, width, height]
	if self.resizer_vlist is not None:
	    self.resizer_vlist.vertices = self._get_resizer_vertices()
	    self.resizer_vlist.colors = color * 40

    def set_scale(self, scale):
	self.scale = scale
	if self.saved_dialog is not None:
	    self.saved_dialog.set_needs_layout()

    def size(self, dialog):
	Control.size(self, dialog)
	self.width = self.texture.width * self.scale
	self.height = self.texture.height * self.scale
	if self.texture_vlist is None:
	    self.texture_group = pyglet.graphics.TextureGroup(
		self.texture, dialog.fg_group)
	    self.texture_vlist = dialog.batch.add(4, gl.GL_QUADS,
                self.texture_group,
                ('v2i', self._get_texture_vertices()),
                ('c4B', (255, 255, 255, 255) * 4),
		('t3f', self.texture.tex_coords))
	if self.resizer_vlist is None:
	    self.resizer_vlist = dialog.batch.add(40, gl.GL_LINES,
                dialog.highlight_group,
                ('v2i', self._get_resizer_vertices()),
                ('c4B', self.color * 40))
	if self.limits_vlist is None:
	    self.limits_vlist = dialog.batch.add(8, gl.GL_LINES,
                dialog.highlight_group,
                ('v2i', self._get_limits_vertices()),
                ('c4B', (192, 192, 192, 255) * 8))




# kytten/frame.py
# Copyrighted (C) 2009 by Conrad "Lynx" Wong

# Classes which wrap one Widget.

# Wrapper: a base class for Widgets which contain one other Widget.
# Frame: positions its contained Widget within a graphic, which it stretches
#        to cover the Widget's area, or the space within which it is contained.
# TitleFrame: like Frame, but has a title region on top as well.

from widgets import Widget, Control, Graphic, Label
from layout import HorizontalLayout, VerticalLayout, GetRelativePoint
from layout import VALIGN_BOTTOM, HALIGN_LEFT, HALIGN_CENTER, HALIGN_RIGHT
from layout import ANCHOR_CENTER

class Wrapper(Widget):
    """
    Wrapper is simply a wrapper around a widget.  While the default
    Wrapper does nothing more interesting, subclasses might decorate the
    widget in some fashion, i.e. Panel might place the widget onto a
    panel, or Scrollable might provide scrollbars to let the widget
    be panned about within its display area.
    """
    def __init__(self, content=None,
                 is_expandable=False, anchor=ANCHOR_CENTER, offset=(0, 0)):
        """
        Creates a new Wrapper around an included Widget.

        @param content The Widget to be wrapped.
        """
        Widget.__init__(self)
        self.content = content
        self.expandable = is_expandable
        self.anchor = anchor
        self.content_offset = offset

    def _get_controls(self):
        """Returns Controls contained by the Wrapper."""
        return self.content._get_controls()

    def delete(self):
        """Deletes graphic elements within the Wrapper."""
        if self.content is not None:
            self.content.delete()
        Widget.delete(self)

    def expand(self, width, height):
        if self.content.is_expandable():
            self.content.expand(width, height)
        self.width = width
        self.height = height

    def is_expandable(self):
        return self.expandable

    def layout(self, x, y):
        """
        Assigns a new position to the Wrapper.

        @param x X coordinate of the Wrapper's lower left corner
        @param y Y coordinate of the Wrapper's lower left corner
        """
        Widget.layout(self, x, y)
        if self.content is not None:
            x, y = GetRelativePoint(
                self, self.anchor,
                self.content, self.anchor, self.content_offset)
            self.content.layout(x, y)

    def set(self, dialog, content):
        """
        Sets a new Widget to be contained in the Wrapper.

        @param dialog The Dialog which contains the Wrapper
        @param content The new Widget to be wrapped
        """
        if self.content is not None:
            self.content.delete()
        self.content = content
        dialog.set_needs_layout()

    def size(self, dialog):
        """
        The default Wrapper wraps up its Widget snugly.

        @param dialog The Dialog which contains the Wrapper
        """
        if dialog is None:
            return
        Widget.size(self, dialog)
        if self.content is not None:
            self.content.size(dialog)
            self.width, self.height = self.content.width, self.content.height
        else:
            self.width = self.height = 0

    def teardown(self):
        self.content.teardown()
        self.content = None
        Widget.teardown(self)

class Frame(Wrapper):
    """
    Frame draws an untitled frame which encloses the dialog's content.
    """
    def __init__(self, content=None, path=['frame'], image_name='image',
                 is_expandable=False, anchor=ANCHOR_CENTER,
                 use_bg_group=False):
        """
        Creates a new Frame surrounding a widget or layout.
        """
        Wrapper.__init__(self, content,
                         is_expandable=is_expandable, anchor=anchor)
        self.frame = None
        self.path = path
        self.image_name = image_name
        self.use_bg_group = use_bg_group

    def delete(self):
        """
        Removes the Frame's graphical elements.
        """
        if self.frame is not None:
            self.frame.delete()
            self.frame = None
        Wrapper.delete(self)

    def expand(self, width, height):
        if self.content.is_expandable():
            content_width, content_height = \
                         self.frame.get_content_size(width, height)
            self.content.expand(content_width, content_height)
        self.width, self.height = width, height

    def layout(self, x, y):
        """
        Positions the Frame.

        @param x X coordinate of lower left corner
        @param y Y coordinate of lower left corner
        """
        self.x, self.y = x, y
        self.frame.update(x, y, self.width, self.height)

        # In some cases the frame graphic element may allocate more space for
        # the content than the content actually fills, due to repeating
        # texture constraints.  Always center the content.
        x, y, width, height = self.frame.get_content_region()
        interior = Widget(width, height)
        interior.x, interior.y = x, y
        x, y = GetRelativePoint(interior, self.anchor,
                                self.content, self.anchor, self.content_offset)
        self.content.layout(x, y)

    def size(self, dialog):
        """
        Determine minimum size of the Frame.

        @param dialog Dialog which contains the Frame
        """
        if dialog is None:
            return
        Wrapper.size(self, dialog)
        if self.frame is None:
            if self.use_bg_group:
                group = dialog.bg_group
            else:
                group = dialog.panel_group
            template = dialog.theme[self.path][self.image_name]
            self.frame = template.generate(
                dialog.theme[self.path]['gui_color'],
                dialog.batch,
                group)
        self.width, self.height = self.frame.get_needed_size(
            self.content.width, self.content.height)

class TitleFrame(VerticalLayout):
    def __init__(self, title, content):
        VerticalLayout.__init__(self, content=[
                HorizontalLayout([
                    Graphic(path=["titlebar", "left"], is_expandable=True),
                    Frame(Label(title, path=["titlebar"]),
                          path=["titlebar", "center"]),
                    Graphic(path=["titlebar", "right"], is_expandable=True),
                ], align=VALIGN_BOTTOM, padding=0),
                Frame(content, path=["titlebar", "frame"], is_expandable=True),
            ], padding=0)

class SectionHeader(HorizontalLayout):
    def __init__(self, title, align=HALIGN_CENTER):
        if align == HALIGN_LEFT:
            left_expand = False
            right_expand = True
        elif align == HALIGN_CENTER:
            left_expand = True
            right_expand = True
        else:  # HALIGN_RIGHT
            left_expand = True
            right_expand = False

        HorizontalLayout.__init__(self, content=[
                Graphic(path=["section", "left"], is_expandable=left_expand),
                Frame(Label(title, path=["section"]),
                      path=['section', 'center'],
                      use_bg_group=True),
                Graphic(path=["section", "right"], is_expandable=right_expand),
            ], align=VALIGN_BOTTOM, padding=0)

class FoldingSection(Control, VerticalLayout):
    def __init__(self, title, content=None, is_open=True, align=HALIGN_CENTER):
        Control.__init__(self)
        if align == HALIGN_LEFT:
            left_expand = False
            right_expand = True
        elif align == HALIGN_CENTER:
            left_expand = True
            right_expand = True
        else:  # HALIGN_RIGHT
            left_expand = True
            right_expand = False

        self.is_open = is_open
        self.folding_content = content
        self.book = Graphic(self._get_image_path())

        self.header = HorizontalLayout([
            Graphic(path=["section", "left"], is_expandable=left_expand),
            Frame(HorizontalLayout([
                      self.book,
                      Label(title, path=["section"]),
                  ]), path=["section", "center"],
                  use_bg_group=True),
            Graphic(path=["section", "right"], is_expandable=right_expand),
            ], align=VALIGN_BOTTOM, padding=0)
        layout = [self.header]
        if self.is_open:
            layout.append(content)

        VerticalLayout.__init__(self, content=layout, align=align)

    def _get_controls(self):
        return VerticalLayout._get_controls(self) + \
               [(self, self.header.x, self.header.x + self.header.width,
                       self.header.y + self.header.height, self.header.y)]

    def _get_image_path(self):
        if self.is_open:
            return ["section", "opened"]
        else:
            return ["section", "closed"]

    def hit_test(self, x, y):
        return self.header.hit_test(x, y)

    def on_mouse_press(self, x, y, button, modifiers):
        self.is_open = not self.is_open
        self.book.delete()
        self.book.path = self._get_image_path()
        if self.is_open:
            self.add(self.folding_content)
        else:
            self.remove(self.folding_content)
            self.folding_content.delete()

    def teardown(self):
        self.folding_content.teardown()
        self.folding_content = None
        VerticalLayout.teardown(self)


# kytten/dialog.py
# Copyrighted (C) 2009 by Conrad "Lynx" Wong

import pyglet
from pyglet import gl

from widgets import Widget, Control, Label
from button import Button
from frame import Wrapper, Frame
from layout import GetRelativePoint, ANCHOR_CENTER
from layout import VerticalLayout, HorizontalLayout

class DialogEventManager(Control):
    def __init__(self):
        """
        Creates a new event manager for a dialog.

        @param content The Widget which we wrap
        """
        Control.__init__(self)
        self.controls = []
        self.control_areas = {}
        self.control_map = {}
        self.hover = None
        self.focus = None
        self.wheel_hint = None
        self.wheel_target = None

    def get_value(self, id):
        widget = self.get_widget(id)
        if widget is not None:
            return widget.get_value()

    def get_values(self):
        retval = {}
        for widget in self.controls:
            if widget.is_input() and widget.id is not None:
                retval[widget.id] = widget.get_value()
        return retval

    def get_widget(self, id):
        return self.control_map.get(id)

    def hit_control(self, x, y, control):
        left, right, top, bottom = self.control_areas[control]
        if x >= left and x < right and y >= bottom and y < top:
            return control.hit_test(x, y)
        else:
            return False

    def on_key_press(self, symbol, modifiers):
        """
        TAB and ENTER will move us between fields, holding shift will
        reverse the direction of our iteration.  We don't handle ESCAPE.
        Otherwise, we pass keys to our child elements.

        @param symbol Key pressed
        @param modifiers Modifiers for key press
        """
        if symbol in [pyglet.window.key.TAB, pyglet.window.key.ENTER]:
            focusable = [x for x in self.controls
                         if x.is_focusable() and not x.is_disabled()]
            if not focusable:
                return

            if modifiers & pyglet.window.key.MOD_SHIFT:
                dir = -1
            else:
                dir = 1

            if self.focus is not None and self.focus in focusable:
                index = focusable.index(self.focus)
            else:
                index = 0 - dir

            new_focus = focusable[(index + dir) % len(focusable)]
            self.set_focus(new_focus)
            new_focus.ensure_visible()

            # If we hit ENTER, and wrapped back to the first focusable,
            # pass the ENTER back so the Dialog can call its on_enter callback
            if symbol != pyglet.window.key.ENTER or \
               new_focus != focusable[0]:
                return pyglet.event.EVENT_HANDLED

        elif symbol != pyglet.window.key.ESCAPE:
            if self.focus is not None and hasattr(self.focus, 'on_key_press'):
                return self.focus.on_key_press(symbol, modifiers)

    def on_key_release(self, symbol, modifiers):
        """Pass key release events to the focus

        @param symbol Key released
        @param modifiers Modifiers for key released
        """
        if self.focus is not None and hasattr(self.focus, 'on_key_release'):
            return self.focus.on_key_release(symbol, modifiers)

    def on_mouse_drag(self, x, y, dx, dy, buttons, modifiers):
        """
        Handles mouse dragging.  If we have a focus, pass it in.

        @param x X coordinate of mouse
        @param y Y coordinate of mouse
        @param dx Delta X
        @param dy Delta Y
        @param buttons Buttons held while moving
        @param modifiers Modifiers to apply to buttons
        """
        if self.focus is not None:
            self.focus.dispatch_event('on_mouse_drag',
                                      x, y, dx, dy, buttons, modifiers)
            return pyglet.event.EVENT_HANDLED

    def on_mouse_motion(self, x, y, dx, dy):
        """
        Handles mouse motion.  We highlight controls that we are hovering
        over.

        @param x X coordinate of mouse
        @param y Y coordinate of mouse
        @param dx Delta X
        @param dy Delta Y
        """
        if self.hover is not None and not self.hit_control(x, y, self.hover):
            self.hover.dispatch_event('on_mouse_motion', x, y, dx, dy)
        new_hover = None
        for control in self.controls:
            if self.hit_control(x, y, control):
                new_hover = control
                break
        self.set_hover(new_hover)
        if self.hover is not None:
            self.hover.dispatch_event('on_mouse_motion', x, y, dx, dy)

    def on_mouse_press(self, x, y, button, modifiers):
        """
        If the focus is set, and the target lies within the focus, pass the
        message down.  Otherwise, check if we need to assign a new focus.
        If the mouse was pressed within our frame but no control was targeted,
        we may be setting up to drag the Dialog around.

        @param x X coordinate of mouse
        @param y Y coordinate of mouse
        @param button Button pressed
        @param modifiers Modifiers to apply to button
        """
        if self.focus is not None and self.hit_control(x, y, self.focus):
            self.focus.dispatch_event('on_mouse_press',
                                      x, y, button, modifiers)
            return pyglet.event.EVENT_HANDLED
        else:
            if self.hit_test(x, y):
                self.set_focus(self.hover)
                if self.focus is not None:
                    self.focus.dispatch_event('on_mouse_press',
                                              x, y, button, modifiers)
                    return pyglet.event.EVENT_HANDLED
            else:
                self.set_focus(None)

    def on_mouse_release(self, x, y, button, modifiers):
        """
        Button was released.  We pass this along to the focus, then we
        generate an on_mouse_motion to handle changing the highlighted
        Control if necessary.

        @param x X coordinate of mouse
        @param y Y coordinate of mouse
        @param button Button released
        @param modifiers Modifiers to apply to button
        """
        self.is_dragging = False
        if self.focus is not None:
            self.focus.dispatch_event('on_mouse_release',
                                      x, y, button, modifiers)
        DialogEventManager.on_mouse_motion(self, x, y, 0, 0)
        return pyglet.event.EVENT_HANDLED

    def on_mouse_scroll(self, x, y, scroll_x, scroll_y):
        """
        Mousewheel was scrolled.  See if we have a wheel target, or
        failing that, a wheel hint.

        @param x X coordinate of mouse
        @param y Y coordinate of mouse
        @param scroll_x Number of clicks horizontally mouse was moved
        @param scroll_y Number of clicks vertically mouse was moved
        """
        if self.wheel_target is not None and \
           self.wheel_target in self.controls:
            self.wheel_target.dispatch_event('on_mouse_scroll',
                                             x, y, scroll_x, scroll_y)
            return pyglet.event.EVENT_HANDLED
        elif self.wheel_hint is not None and \
             self.wheel_hint in self.controls:
            self.wheel_hint.dispatch_event('on_mouse_scroll',
                                           x, y, scroll_x, scroll_y)
            return pyglet.event.EVENT_HANDLED

    def on_text(self, text):
        if self.focus and text != u'\r':
            try:
                return getattr(self.focus, 'on_text')(text)
            except KeyError:
                return pyglet.event.EVENT_UNHANDLED

    def on_text_motion(self, motion):
        if self.focus:
            try:
                return getattr(self.focus, 'on_text_motion')(motion)
            except KeyError:
                return pyglet.event.EVENT_UNHANDLED

    def on_text_motion_select(self, motion):
        if self.focus:
            try:
                return getattr(self.focus, 'on_text_motion_select')(motion)
            except KeyError:
                return pyglet.event.EVENT_UNHANDLED

    def on_update(self, dt):
        """
        We update our layout only when it's time to construct another frame.
        Since we may receive several resize events within this time, this
        ensures we don't resize too often.

        @param dialog The Dialog containing the controls
        @param dt Time passed since last update event (in seconds)
        """
        for control in self.controls:
            control.dispatch_event('on_update', dt)

    def set_focus(self, focus):
        """
        Sets a new focus, dispatching lose and gain focus events appropriately

        @param focus The new focus, or None if no focus
        """
        if self.focus == focus:
            return
        if self.focus is not None:
            self.focus.dispatch_event('on_lose_focus')
        self.focus = focus
        if focus is not None:
            focus.dispatch_event('on_gain_focus')

    def set_hover(self, hover):
        """
        Sets a new highlight, dispatching lose and gain highlight events
        appropriately

        @param hover The new highlight, or None if no highlight
        """
        if self.hover == hover:
            return
        if self.hover is not None:
            self.hover.dispatch_event('on_lose_highlight')
        self.hover = hover
        if hover is not None:
            hover.dispatch_event('on_gain_highlight')

    def set_wheel_hint(self, control):
        self.wheel_hint = control

    def set_wheel_target(self, control):
        self.wheel_target = control

    def teardown(self):
        self.controls = []
        self.control_map = {}
        self.focus = None
        self.hover = None
        self.wheel_hint = None
        self.wheel_target = None

    def update_controls(self):
        """Update our list of controls which may respond to user input."""
        controls = self._get_controls()
        self.controls = []
        self.control_areas = {}
        self.control_map = {}
        for control, left, right, top, bottom in controls:
            self.controls.append(control)
            self.control_areas[control] = (left, right, top, bottom)
            if control.id is not None:
                self.control_map[control.id] = control

        if self.hover is not None and self.hover not in self.controls:
            self.set_hover(None)
        if self.focus is not None and self.focus not in self.controls:
            self.set_focus(None)

kytten_next_dialog_order_id = 0
def GetNextDialogOrderId():
    global kytten_next_dialog_order_id
    kytten_next_dialog_order_id += 1
    return kytten_next_dialog_order_id

class DialogGroup(pyglet.graphics.OrderedGroup):
    """
    Ensure that all Widgets within a Dialog can be drawn with
    blending enabled, and that our Dialog will be drawn in a particular
    order relative to other Dialogs.
    """
    def __init__(self, parent=None):
        """
        Creates a new DialogGroup.  By default we'll be on top.

        @param parent Parent group
        """
        pyglet.graphics.OrderedGroup.__init__(
            self, GetNextDialogOrderId(), parent)
        self.real_order = self.order

    def __cmp__(self, other):
        """
        When compared with other DialogGroups, we'll return our real order
        compared against theirs; otherwise use the OrderedGroup comparison.
        """
        if isinstance(other, DialogGroup):
            return cmp(self.real_order, other.real_order)
        else:
            return OrderedGroup.__cmp__(self, other)

    def is_on_top(self):
        """
        Are we the top dialog group?
        """
        global kytten_next_dialog_order_id
        return self.real_order == kytten_next_dialog_order_id

    def pop_to_top(self):
        """
        Put us on top of other dialog groups.
        """
        self.real_order = GetNextDialogOrderId()

    def set_state(self):
        """
        Ensure that blending is set.
        """
        gl.glPushAttrib(gl.GL_ENABLE_BIT | gl.GL_CURRENT_BIT)
        gl.glEnable(gl.GL_BLEND)
        gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)

    def unset_state(self):
        """
        Restore previous blending state.
        """
        gl.glPopAttrib()

class Dialog(Wrapper, DialogEventManager):
    """
    Defines a new GUI.  By default it can contain only one element, but that
    element can be a Layout of some kind which can contain multiple elements.
    Pass a Theme in to set the graphic appearance of the Dialog.

    The Dialog is always repositioned in relationship to the window, and
    handles resize events accordingly.
    """
    def __init__(self, content=None, window=None, batch=None, group=None,
                 anchor=ANCHOR_CENTER, offset=(0, 0), parent=None,
                 theme=None, movable=True, on_enter=None, on_escape=None):
        """
        Creates a new dialog.

        @param content The Widget which we wrap
        @param window The window to which we belong; used to set the
                      mouse cursor when appropriate.  If set, we will
                      add ourself to the window as a handler.
        @param batch Batch in which we are to place our graphic elements;
                     may be None if we are to create our own Batch
        @param group Group in which we are to place our graphic elements;
                     may be None
        @param anchor Anchor point of the window, relative to which we
                      are positioned.  If ANCHOR_TOP_LEFT is specified,
                      our top left corner will be aligned to the window's
                      top left corner; if ANCHOR_CENTER is specified,
                      our center will be aligned to the window's center,
                      and so forth.
        @param offset Offset from the anchor point.  A positive X is always
                      to the right, a positive Y to the upward direction.
        @param theme The Theme which we are to use to generate our graphical
                     appearance.
        @param movable True if the dialog is able to be moved
        @param on_enter Callback for when user presses enter on the last
                        input within this dialog, i.e. form submit
        @param on_escape Callback for when user presses escape
        """
        assert isinstance(theme, dict)
        Wrapper.__init__(self, content=content)
        DialogEventManager.__init__(self)

        self.window = window
        self.anchor = anchor
        self.offset = offset
        self.theme = theme
        self.is_movable = movable
        self.on_enter = on_enter
        self.on_escape = on_escape
        if batch is None:
            self.batch = pyglet.graphics.Batch()
            self.own_batch = True
        else:
            self.batch = batch
            self.own_batch = False
        self.root_group = DialogGroup(parent=group)
        self.panel_group = pyglet.graphics.OrderedGroup(0, self.root_group)
        self.bg_group = pyglet.graphics.OrderedGroup(1, self.root_group)
        self.fg_group = pyglet.graphics.OrderedGroup(2, self.root_group)
        self.highlight_group = pyglet.graphics.OrderedGroup(3, self.root_group)
        self.needs_layout = True
        self.is_dragging = False

        if window is None:
            self.screen = Widget()
        else:
            width, height = window.get_size()
            self.screen = Widget(width=width, height=height)
            window.push_handlers(self)

    def do_layout(self):
        """
        We lay out the Dialog by first determining the size of all its
        chlid Widgets, then laying ourself out relative to the parent window.
        """
        # Determine size of all components
        self.size(self)

        # Calculate our position relative to our containing window,
        # making sure that we fit completely on the window.  If our offset
        # would send us off the screen, constrain it.
        x, y = GetRelativePoint(self.screen, self.anchor,
                                self, None, (0, 0))
        max_offset_x = self.screen.width - self.width - x
        max_offset_y = self.screen.height - self.height - y
        offset_x, offset_y = self.offset
        offset_x = max(min(offset_x, max_offset_x), -x)
        offset_y = max(min(offset_y, max_offset_y), -y)
        self.offset = (offset_x, offset_y)
        x += offset_x
        y += offset_y

        # Perform the actual layout now!
        self.layout(x, y)
        self.update_controls()

        self.needs_layout = False

    def draw(self):
        assert self.own_batch
        self.batch.draw()

    def ensure_visible(self, control):
        """
        Ensure a control is visible.  For Dialog, this doesn't matter
        since we don't scroll.
        """
        pass

    def get_root(self):
        return self

    def on_key_press(self, symbol, modifiers):
        """
        We intercept TAB, ENTER, and ESCAPE events.  TAB and ENTER will
        move us between fields, holding shift will reverse the direction
        of our iteration.  ESCAPE may cause us to send an on_escape
        callback.

        Otherwise, we pass key presses to our child elements.

        @param symbol Key pressed
        @param modifiers Modifiers for key press
        """
        retval = DialogEventManager.on_key_press(self, symbol, modifiers)
        if not retval:
            if symbol in [pyglet.window.key.TAB, pyglet.window.key.ENTER]:
                if self.on_enter is not None:
                    self.on_enter(self)
                    return pyglet.event.EVENT_HANDLED
            elif symbol == pyglet.window.key.ESCAPE:
                if self.on_escape is not None:
                    self.on_escape(self)
                    return pyglet.event.EVENT_HANDLED
        return retval

    def on_mouse_drag(self, x, y, dx, dy, buttons, modifiers):
        """
        Handles mouse dragging.  If we have a focus, pass it in.  Otherwise
        if we are movable, and we were being dragged, move the window.

        @param x X coordinate of mouse
        @param y Y coordinate of mouse
        @param dx Delta X
        @param dy Delta Y
        @param buttons Buttons held while moving
        @param modifiers Modifiers to apply to buttons
        """
        if not DialogEventManager.on_mouse_drag(self, x, y, dx, dy,
                                                buttons, modifiers):
            if self.is_movable and self.is_dragging:
                x, y = self.offset
                self.offset = (int(x + dx), int(y + dy))
                self.set_needs_layout()
                return pyglet.event.EVENT_HANDLED

    def on_mouse_press(self, x, y, button, modifiers):
        """
        If the focus is set, and the target lies within the focus, pass the
        message down.  Otherwise, check if we need to assign a new focus.
        If the mouse was pressed within our frame but no control was targeted,
        we may be setting up to drag the Dialog around.

        @param x X coordinate of mouse
        @param y Y coordinate of mouse
        @param button Button pressed
        @param modifiers Modifiers to apply to button
        """
        retval = DialogEventManager.on_mouse_press(self, x, y,
                                             button, modifiers)
        if self.hit_test(x, y):
            if not self.root_group.is_on_top():
                self.pop_to_top()
            if not retval:
                self.is_dragging = True
                retval = pyglet.event.EVENT_HANDLED
        return retval

    def on_mouse_release(self, x, y, button, modifiers):
        """
        Button was released.  We pass this along to the focus, then we
        generate an on_mouse_motion to handle changing the highlighted
        Control if necessary.

        @param x X coordinate of mouse
        @param y Y coordinate of mouse
        @param button Button released
        @param modifiers Modifiers to apply to button
        """
        self.is_dragging = False
        return DialogEventManager.on_mouse_release(self, x, y,
                                                   button, modifiers)

    def on_resize(self, width, height):
        """
        Update our knowledge of the window's width and height.

        @param width Width of the window
        @param height Height of the window
        """
        if self.screen.width != width or self.screen.height != height:
            self.screen.width, self.screen.height = width, height
            self.needs_layout = True

    def on_update(self, dt):
        """
        We update our layout only when it's time to construct another frame.
        Since we may receive several resize events within this time, this
        ensures we don't resize too often.

        @param dt Time passed since last update event (in seconds)
        """
        if self.needs_layout:
            self.do_layout()
        DialogEventManager.on_update(self, dt)

    def pop_to_top(self):
        """
        Pop our dialog group to the top, and force our batch to re-sort
        the groups.  Also, puts our event handler on top of the window's
        event handler stack.
        """
        self.root_group.pop_to_top()
        self.batch._draw_list_dirty = True  # forces resorting groups
        if self.window is not None:
            self.window.remove_handlers(self)
            self.window.push_handlers(self)

    def set_needs_layout(self):
        """
        True if we should redo the Dialog layout on our next update.
        """
        self.needs_layout = True

    def teardown(self):
        DialogEventManager.teardown(self)
        if self.content is not None:
            self.content.teardown()
            self.content = None
        if self.window is not None:
            self.window.remove_handlers(self)
            self.window = None
        self.batch._draw_list_dirty = True  # forces resorting groups

class PopupMessage(Dialog):
    """A simple fire-and-forget dialog."""

    def __init__(self, text="", window=None, batch=None, group=None,
                 theme=None, on_escape=None):
        def on_ok(dialog=None):
            if on_escape is not None:
                on_escape(self)
            self.teardown()

        return Dialog.__init__(self, content=Frame(
            VerticalLayout([
                Label(text),
                Button("Ok", on_click=on_ok),
            ])),
            window=window, batch=batch, group=group,
            theme=theme, movable=True,
            on_enter=on_ok, on_escape=on_ok)

class PopupConfirm(Dialog):
    """An ok/cancel-style dialog.  Escape defaults to cancel."""

    def __init__(self, text="", ok="Ok", cancel="Cancel",
                 window=None, batch=None, group=None, theme=None,
                 on_ok=None, on_cancel=None):
        def on_ok_click(dialog=None):
            if on_ok is not None:
                on_ok(self)
            self.teardown()

        def on_cancel_click(dialog=None):
            if on_cancel is not None:
                on_cancel(self)
            self.teardown()

        return Dialog.__init__(self, content=Frame(
            VerticalLayout([
                Label(text),
                HorizontalLayout([
                    Button(ok, on_click=on_ok_click),
                    None,
                    Button(cancel, on_click=on_cancel_click)
                ]),
            ])),
            window=window, batch=batch, group=group,
            theme=theme, movable=True,
            on_enter=on_ok_click, on_escape=on_cancel_click)


# kytten/file_dialogs.py
# Copyrighted (C) 2009 by Conrad "Lynx" Wong

import glob
import os
import pyglet
from pyglet import gl

from button import Button
from dialog import Dialog
from frame import Frame, SectionHeader
from layout import VerticalLayout, HorizontalLayout
from layout import ANCHOR_CENTER, HALIGN_LEFT, VALIGN_BOTTOM
from menu import Menu, Dropdown
from scrollable import Scrollable
from text_input import Input
from widgets import Label

class FileLoadDialog(Dialog):
    def __init__(self, path=os.getcwd(), extensions=[], title="Select File",
                 width=540, height=300, window=None, batch=None, group=None,
                 anchor=ANCHOR_CENTER, offset=(0, 0),
                 theme=None, movable=True, on_select=None, on_escape=None):
        self.path = path
        self.extensions = extensions
        self.title = title
        self.on_select = on_select
        self.selected_file = None
        self._set_files()

        def on_parent_menu_select(choice):
            self._select_file(self.parents_dict[choice])

        def on_menu_select(choice):
            self._select_file(self.files_dict[choice])

        self.dropdown = Dropdown(options=self.parents,
                                 selected=self.parents[-1],
                                 align=VALIGN_BOTTOM,
                                 on_select=on_parent_menu_select)
        self.menu = Menu(options=self.files, align=HALIGN_LEFT,
                         on_select=on_menu_select)
        self.scrollable = Scrollable(
            VerticalLayout([self.dropdown, self.menu], align=HALIGN_LEFT),
            width=width, height=height)

        content = self._get_content()
        Dialog.__init__(self, content, window=window, batch=batch, group=group,
                        anchor=anchor, offset=offset, theme=theme,
                        movable=movable, on_escape=on_escape)

    def _get_content(self):
        return Frame(
            VerticalLayout([
                SectionHeader(self.title),
                self.scrollable,
            ], align=HALIGN_LEFT)
        )

    def _select_file(self, filename):
        if os.path.isdir(filename):
            self.path = filename
            self._set_files()
            self.dropdown.set_options(self.parents,
                                      selected=self.parents[-1])
            self.menu.set_options(self.files)
        else:
            self.selected_file = filename
            if self.on_select is not None:
                self.on_select(filename)

    def _set_files(self):
        # Once we have a new path, update our files
        filenames = glob.glob(os.path.join(self.path, '*'))

        # First, a list of directories
        self.parents = []
        self.parents_dict = {}
        path = self.path
        index = 1
        while 1:
            name = "%d %s" % (index, os.path.basename(path) or path)
            self.parents_dict[name] = path
            self.parents.append(name)
            index += 1
            path, child = os.path.split(path)
            if not child:
                break
        self.parents.reverse()

        files = [("%s (dir)" % os.path.basename(x), x) for x in filenames
                 if os.path.isdir(x)]

        # Now add the files that match the extensions
        if self.extensions:
            for filename in filenames:
                if os.path.isfile(filename):
                    ext = os.path.splitext(filename)[1]
                    if ext in self.extensions:
                        files.append((os.path.basename(filename), filename))
        else:
            files.extend([(os.path.basename(x), x) for x in filenames
                          if os.path.isfile(x)])

        self.selected_file = None
        self.files_dict = dict(files)
        self.files = self.files_dict.keys()

        def dir_sort(x, y):
            if x.endswith(' (dir)') and y.endswith(' (dir)'):
                return cmp(x, y)
            elif x.endswith(' (dir)'):
                return -1
            elif y.endswith(' (dir)'):
                return 1
            else:
                return cmp(x, y)
        self.files.sort(dir_sort)

    def get(self):
        return self.selected_file

    def size(self, dialog):
        Dialog.size(self, dialog)

    def teardown(self):
        self.on_select = None
        Dialog.teardown(self)

class FileSaveDialog(FileLoadDialog):
    def __init__(self, *args, **kwargs):
        self.text_input = Input()

        # Set up buttons to be shown in our contents
        def on_save():
            self._do_select()
        self.save_button = Button("Save", on_click=on_save)

        def on_cancel():
            self._do_cancel()
        self.cancel_button = Button("Cancel", on_click=on_cancel)

        FileLoadDialog.__init__(self, *args, **kwargs)

        # Setup our event handlers
        def on_enter(dialog):
            self._do_select()
        self.on_enter = on_enter

        self.real_on_select = self.on_select
        def on_select(filename):
            self.text_input.set_text(filename)
        self.on_select = on_select

    def _do_cancel(self):
        if self.on_escape is not None:
            self.on_escape(self)
        else:
            self.delete()
            self.window.remove_handlers(self)

    def _do_select(self):
        filename = self.text_input.get_text()
        path, base = os.path.split(filename)
        if not base:
            filename = None
        elif not path:
            filename = os.path.join(self.path, filename)
        if self.real_on_select is not None:
            if self.id is not None:
                self.real_on_select(self.id, filename)
            else:
                self.real_on_select(filename)

    def _get_content(self):
        return Frame(
            VerticalLayout([
                SectionHeader(self.title),
                self.scrollable,
                Label("Filename:"),
                self.text_input,
                HorizontalLayout([
                    self.save_button, None, self.cancel_button
                ]),
            ], align=HALIGN_LEFT)
        )

class DirectorySelectDialog(FileLoadDialog):
    def __init__(self, *args, **kwargs):
        self.text_input = Input()

        # Set up buttons to be shown in our contents
        def on_select_button():
            self._do_select()
        self.select_button = Button("Select", on_click=on_select_button)

        def on_cancel_button():
            self._do_cancel()
        self.cancel_button = Button("Cancel", on_click=on_cancel_button)

        FileLoadDialog.__init__(self, *args, **kwargs)

        # Setup our event handlers
        def on_enter(dialog):
            self._do_select()
        self.on_enter = on_enter

        self.real_on_select = self.on_select
        def on_select(filename):
            self.text_input.set_text(filename)
        self.on_select = on_select

        def on_parent_menu_select(choice):
            self.text_input.set_text(self.parents_dict[choice])
            self._do_open()
        self.dropdown.on_select = on_parent_menu_select

    def _do_cancel(self):
        if self.on_escape is not None:
            self.on_escape(self)
        else:
            self.delete()
            self.window.remove_handlers(self)

    def _do_open(self):
        filename = self.text_input.get_text()
        if os.path.isdir(filename):
            self.path = filename
            self._set_files()
            self.dropdown.set_options(self.parents,
                                      selected=self.parents[-1])
            self.menu.set_options(self.files)

    def _do_select(self):
        filename = self.text_input.get_text()
        path, base = os.path.split(filename)
        if not base:
            filename = None
        elif not path:
            filename = os.path.join(self.path, filename)
        if self.real_on_select is not None:
            if self.id is not None:
                self.real_on_select(self.id, filename)
            else:
                self.real_on_select(filename)

    def _get_content(self):
        return Frame(
            VerticalLayout([
                SectionHeader(self.title),
                self.scrollable,
                Label("Directory:"),
                self.text_input,
                HorizontalLayout([
                    self.select_button, None, self.cancel_button
                ]),
            ], align=HALIGN_LEFT)
        )

    def _select_file(self, filename):
        if not os.path.isdir(filename):
            return  # we accept only directories!

        if self.selected_file == filename:
            if filename != self.path:
                self._do_open()
            else:
                self._do_select()
        else:
            self.selected_file = filename
            if self.on_select is not None:
                self.on_select(filename)

    def _set_files(self):
        # Once we have a new path, update our files
        filenames = glob.glob(os.path.join(self.path, '*'))

        # First, a list of directories
        self.parents = []
        self.parents_dict = {}
        path = self.path
        index = 1
        while 1:
            name = "%d %s" % (index, os.path.basename(path) or path)
            self.parents_dict[name] = path
            self.parents.append(name)
            index += 1
            path, child = os.path.split(path)
            if not child:
                break
        self.parents.reverse()

        files = [('(this dir)', self.path)] + \
                [("%s (dir)" % os.path.basename(x), x) for x in filenames
                 if os.path.isdir(x)]
        # Now add the files that match the extensions
        if self.extensions:
            for filename in filenames:
                if os.path.isfile(filename):
                    ext = os.path.splitext(filename)[1]
                    if ext in self.extensions:
                        files.append(('-%s' % os.path.basename(filename),
                                      filename))
        else:
            files.extend([('-%s' % os.path.basename(x), x) for x in filenames
                          if os.path.isfile(x)])

        self.selected_file = None
        self.files_dict = dict(files)
        self.files = self.files_dict.keys()

        def dir_sort(x, y):
            if x == '(this dir)':
                return -1
            elif x.endswith(' (dir)') and y.endswith(' (dir)'):
                return cmp(x, y)
            elif x.endswith(' (dir)') and y != '(this dir)':
                return -1
            elif y.endswith(' (dir)'):
                return 1
            else:
                return cmp(x, y)
        self.files.sort(dir_sort)



# kytten/button.py
# Copyrighted (C) 2009 by Conrad "Lynx" Wong

import pyglet
from widgets import Control
from override import KyttenLabel

class Button(Control):
    """
    A simple text-labeled button.
    """
    def __init__(self, text="", id=None, on_click=None, disabled=False):
        """
        Creates a new Button.  The provided text will be used to caption the
        button.

        @param text Label for the button
        @param on_click Callback for the button
        @param disabled True if the button should be disabled
        """
        Control.__init__(self, id=id, disabled=disabled)
        self.text = text
        self.on_click = on_click
        self.label = None
        self.button = None
        self.highlight = None
        self.is_pressed = False

    def delete(self):
        """
        Clean up our graphic elements
        """
        Control.delete(self)
        if self.button is not None:
            self.button.delete()
            self.button = None
        if self.label is not None:
            self.label.delete()
            self.label = None
        if self.highlight is not None:
            self.highlight.delete()
            self.highlight = None

    def layout(self, x, y):
        """
        Places the Button.

        @param x X coordinate of lower left corner
        @param y Y coordinate of lower left corner
        """
        Control.layout(self, x, y)
        self.button.update(self.x, self.y, self.width, self.height)
        if self.highlight is not None:
            self.highlight.update(self.x, self.y, self.width, self.height)
        x, y, width, height = self.button.get_content_region()
        font = self.label.document.get_font()
        self.label.x = x + width/2 - self.label.content_width/2
        self.label.y = y + height/2 - font.ascent/2 - font.descent

    def on_gain_highlight(self):
        Control.on_gain_highlight(self)
        self.size(self.saved_dialog)
        if self.highlight is not None:
            self.highlight.update(self.x, self.y, self.width, self.height)

    def on_lose_highlight(self):
        Control.on_lose_highlight(self)
        if self.highlight is not None:
            self.highlight.delete()
            self.highlight = None

    def on_mouse_press(self, x, y, button, modifiers):
        if not self.is_pressed and not self.is_disabled():
            self.is_pressed = True

            # Delete the button to force it to be redrawn
            self.delete()
            self.saved_dialog.set_needs_layout()

    def on_mouse_release(self, x, y, button, modifiers):
        if self.is_pressed:
            self.is_pressed = False

            # Delete the button to force it to be redrawn
            self.delete()
            self.saved_dialog.set_needs_layout()

            # Now, if mouse is still inside us, signal on_click
            if self.on_click is not None and self.hit_test(x, y):
                if self.id is not None:
                    self.on_click(self.id)
                else:
                    self.on_click()

    def size(self, dialog):
        """
        Sizes the Button.  If necessary, creates the graphic elements.

        @param dialog Dialog which contains the Button
        """
        if dialog is None:
            return
        Control.size(self, dialog)
        if self.is_pressed:
            path = ['button', 'down']
        else:
            path = ['button', 'up']
        if self.is_disabled():
            color = dialog.theme[path]['disabled_color']
        else:
            color = dialog.theme[path]['gui_color']
        if self.button is None:
            self.button = dialog.theme[path]['image'].generate(
                color,
                dialog.batch, dialog.bg_group)
        if self.highlight is None and self.is_highlight():
            self.highlight = dialog.theme[path]['highlight']['image'].\
                generate(dialog.theme[path]['highlight_color'],
                         dialog.batch,
                         dialog.bg_group)
        if self.label is None:
            self.label = KyttenLabel(self.text,
                font_name=dialog.theme[path]['font'],
                font_size=dialog.theme[path]['font_size'],
                color=dialog.theme[path]['text_color'],
                batch=dialog.batch, group=dialog.fg_group)

        # Treat the height of the label as ascent + descent
        font = self.label.document.get_font()
        height = font.ascent - font.descent # descent is negative
        self.width, self.height = self.button.get_needed_size(
            self.label.content_width, height)

    def teardown(self):
        self.on_click = None
        Control.teardown(self)


# kytten/slider.py
# Copyrighted (C) 2009 by Conrad "Lynx" Wong

import pyglet
from widgets import Control

class Slider(Control):
    """
    A horizontal slider.  Position is measured from 0.0 to 1.0.
    """
    IMAGE_BAR = ['slider', 'bar']
    IMAGE_KNOB = ['slider', 'knob']
    IMAGE_STEP = ['slider', 'step']

    def __init__(self, value=0.0, min_value=0.0, max_value=1.0, steps=None,
                 width=100, id=None, on_set=None, disabled=False):
        """
        Creates a new slider.

        @param min_value Minimum value
        @param max_value Maximum value
        @param steps None if this slider should cover the range from 0.0 to
                     1.0 smoothly, otherwise we will divide the range
                     up into steps.  For instance, 2 steps would give the
                     possible values 0, 0.5, and 1.0 (then multiplied by scale)
        @param width Minimum width of the tracking area.  Note that this
                     is the interior length of the slider, not the overall
                     size.
        @param id ID for identifying this slider.
        @param on_set Callback function for when the value of this slider
                      changes.
        @param diasbled True if the slider should be disabled
        """
        Control.__init__(self, id=id, disabled=disabled)
        self.min_value = min_value
        self.max_value = max_value
        self.steps = steps
        self.min_width = width
        self.on_set = on_set
        self.bar = None
        self.knob = None
        self.markers = []
        self.pos = max(
            min(float(value - min_value) / (max_value - min_value), 1.0),
            0.0)
        self.offset = (0, 0)
        self.step_offset = (0, 0)
        self.padding = (0, 0, 0, 0)
        self.is_dragging = False

    def delete(self):
        """
        Delete all graphic elements used by the slider
        """
        if self.bar is not None:
            self.bar.delete()
            self.bar = None
        if self.knob is not None:
            self.knob.delete()
            self.knob = None
        for marker in self.markers:
            marker.delete()
        self.markers = []

    def expand(self, width, height):
        self.width = width

    def get_value(self):
        return self.min_value + (self.max_value - self.min_value) * self.pos

    def is_expandable(self):
        return True

    def is_input(self):
        return True

    def layout(self, x, y):
        """
        Lays out the slider components

        @param x X coordinate of lower left corner
        @param y Y coordinate of lower left corner
        """
        self.x, self.y = x, y
        if self.bar is not None:
            left, right, top, bottom = self.padding
            self.bar.update(x + left, y + bottom,
                            self.width - left - right,
                            self.height - top - bottom)
            x, y, width, height = self.bar.get_content_region()

            if self.knob is not None:
                offset_x, offset_y = self.offset
                self.knob.update(x + int(width * self.pos) + offset_x,
                                 y + offset_y,
                                 self.knob.width, self.knob.height)

            if self.markers:
                step = float(width) / self.steps
                offset_x, offset_y = self.step_offset
                for n in xrange(0, self.steps + 1):
                    self.markers[n].update(int(x + step * n) + offset_x,
                                           y + offset_y,
                                           self.markers[n].width,
                                           self.markers[n].height)

    def on_mouse_drag(self, x, y, dx, dy, buttons, modifiers):
        if self.is_dragging and self.bar is not None:
            bar_x, bar_y, bar_width, bar_height = self.bar.get_content_region()
            self.set_pos(self.pos + float(dx) / bar_width)

    def on_mouse_press(self, x, y, button, modifiers):
        if self.is_disabled():
            return

        self.is_dragging = True
        if self.bar is not None:
            bar_x, bar_y, bar_width, bar_height = self.bar.get_content_region()
            self.set_pos(float(x - bar_x) / bar_width)

    def on_mouse_release(self, x, y, button, modifiers):
        if self.is_disabled():
            return

        self.is_dragging = False
        self.snap_to_nearest()
        if self.on_set is not None:
            value = self.get_value()
            if self.id is not None:
                self.on_set(self.id, value)
            else:
                self.on_set(value)

    def set_pos(self, pos):
        self.pos = max(min(pos, 1.0), 0.0)
        if self.bar is not None and self.knob is not None:
            x, y, width, height = self.bar.get_content_region()
            offset_x, offset_y = self.offset
            self.knob.update(x + int(width * self.pos) + offset_x,
                             y + offset_y,
                             self.knob.width, self.knob.height)

    def size(self, dialog):
        """
        Creates slider components.
        """
        if dialog is None:
            return
        Control.size(self, dialog)
        if self.is_disabled():
            color = dialog.theme['slider']['disabled_color']
        else:
            color = dialog.theme['slider']['gui_color']
        if self.bar is None:
            path = self.IMAGE_BAR
            self.bar = dialog.theme[path]['image'].generate(
                color,
                dialog.batch, dialog.bg_group)
            self.padding = dialog.theme[path]['padding']
        if self.knob is None:
            path = self.IMAGE_KNOB
            self.knob = dialog.theme[path]['image'].generate(
                color,
                dialog.batch, dialog.highlight_group)
            self.offset = dialog.theme[path]['offset']
        if not self.markers and self.steps is not None:
            path = self.IMAGE_STEP
            for n in xrange(0, self.steps + 1):
                self.markers.append(
                    dialog.theme[path]['image'].generate(
                        color,
                        dialog.batch, dialog.fg_group))
            self.step_offset = dialog.theme[path]['offset']
        width, height = self.bar.get_needed_size(self.min_width, 0)
        left, right, top, bottom = self.padding
        self.width = width + left + right
        self.height = height + top + bottom

    def snap_to_nearest(self):
        if self.steps is not None:
            n = int(self.pos * self.steps + 0.5)
            self.set_pos(float(n) / self.steps)

    def teardown(self):
        self.on_set = None
        Control.teardown(self)


# safe_eval
# Copyrighted (C) Michael Spencer
#
# Source: http://code.activestate.com/recipes/364469/

import compiler

class Unsafe_Source_Error(Exception):
    def __init__(self,error,descr = None,node = None):
	self.error = error
	self.descr = descr
	self.node = node
	self.lineno = getattr(node,"lineno",None)

    def __repr__(self):
	return "Line %d.  %s: %s" % (self.lineno, self.error, self.descr)
    __str__ = __repr__

class SafeEval(object):

    def visit(self, node,**kw):
	cls = node.__class__
	meth = getattr(self,'visit'+cls.__name__,self.default)
	return meth(node, **kw)

    def default(self, node, **kw):
	for child in node.getChildNodes():
	    return self.visit(child, **kw)

    visitExpression = default

    def visitConst(self, node, **kw):
	return node.value

    def visitDict(self, node,**kw):
	return dict([(self.visit(k),self.visit(v)) for k,v in node.items])

    def visitTuple(self, node, **kw):
	return tuple(self.visit(i) for i in node.nodes)

    def visitList(self, node, **kw):
	return [self.visit(i) for i in node.nodes]

    def visitUnarySub(self, node, **kw):
	return -self.visit(node.expr)

class SafeEvalWithErrors(SafeEval):

    def default(self, node, **kw):
	raise Unsafe_Source_Error("Unsupported source construct",
				node.__class__,node)

    def visitName(self,node, **kw):
	raise Unsafe_Source_Error("Strings must be quoted",
				 node.name, node)

    # Add more specific errors if desired


def safe_eval(source, fail_on_error = True):
    walker = fail_on_error and SafeEvalWithErrors() or SafeEval()
    try:
	ast = compiler.parse(source,"eval")
    except SyntaxError, err:
	raise Unsafe_Source_Error(err, source)
    try:
	return walker.visit(ast)
    except Unsafe_Source_Error, err:
	raise


# kytten/layout.py
# Copyrighted (C) 2009 by Conrad "Lynx" Wong

# Layouts are arrangements of multiple Widgets.
#
# VerticalLayout: a stack of Widgets, one on top of another.
# HorizontalLayout: a row of Widgets, side by side.
# GridLayout: a table of Widgets.
# FreeLayout: an open area within which Widgets may be positioned freely,
#             relative to one of its anchor points.

import pyglet
from pyglet import gl

from widgets import Widget, Control, Spacer, Graphic, Label

# GUI layout constants

VALIGN_TOP = 1
VALIGN_CENTER = 0
VALIGN_BOTTOM = -1

HALIGN_LEFT = -1
HALIGN_CENTER = 0
HALIGN_RIGHT = 1

ANCHOR_TOP_LEFT = (VALIGN_TOP, HALIGN_LEFT)
ANCHOR_TOP = (VALIGN_TOP, HALIGN_CENTER)
ANCHOR_TOP_RIGHT = (VALIGN_TOP, HALIGN_RIGHT)
ANCHOR_LEFT = (VALIGN_CENTER, HALIGN_LEFT)
ANCHOR_CENTER = (VALIGN_CENTER, HALIGN_CENTER)
ANCHOR_RIGHT = (VALIGN_CENTER, HALIGN_RIGHT)
ANCHOR_BOTTOM_LEFT = (VALIGN_BOTTOM, HALIGN_LEFT)
ANCHOR_BOTTOM = (VALIGN_BOTTOM, HALIGN_CENTER)
ANCHOR_BOTTOM_RIGHT = (VALIGN_BOTTOM, HALIGN_RIGHT)

def GetRelativePoint(parent, parent_anchor, child, child_anchor, offset):
    valign, halign = parent_anchor or ANCHOR_CENTER

    if valign == VALIGN_TOP:
        y = parent.y + parent.height
    elif valign == VALIGN_CENTER:
        y = parent.y + parent.height / 2
    else: # VALIGN_BOTTOM
        y = parent.y

    if halign == HALIGN_LEFT:
        x = parent.x
    elif halign == HALIGN_CENTER:
        x = parent.x + parent.width / 2
    else: # HALIGN_RIGHT
        x = parent.x + parent.width

    valign, halign = child_anchor or (valign, halign)
    offset_x, offset_y = offset

    if valign == VALIGN_TOP:
        y += offset_y - child.height
    elif valign == VALIGN_CENTER:
        y += offset_y - child.height/2
    else: # VALIGN_BOTTOM
        y += offset_y

    if halign == HALIGN_LEFT:
        x += offset_x
    elif halign == HALIGN_CENTER:
        x += offset_x - child.width / 2
    else: # HALIGN_RIGHT
        x += offset_x - child.width

    return (x, y)

class VerticalLayout(Widget):
    """
    Arranges Widgets on top of each other, from top to bottom.
    """
    def __init__(self, content=[], align=HALIGN_CENTER, padding=5):
        """
        Creates a new VerticalLayout.

        @param content A list of Widgets to be arranged
        @param align HALIGN_LEFT if Widgets are to be left-justified,
                     HALIGN_CENTER if they should be centered, and
                     HALIGN_RIGHT if they are to be right-justified.
        @param padding This amount of padding is inserted between widgets.
        """
        assert isinstance(content, list) or isinstance(content, tuple)
        Widget.__init__(self)
        self.align = align
        self.padding = padding
        self.content = [x or Spacer() for x in content]
        self.expandable = []

    def _get_controls(self):
        """
        Returns Controls within the layout.
        """
        controls = []
        for item in self.content:
            controls += item._get_controls()
        return controls

    def add(self, item):
        """
        Adds a new Widget to the layout.

        @param item The Widget to be added
        """
        self.content.append(item or Spacer())
        self.saved_dialog.set_needs_layout()

    def delete(self):
        """Deletes all graphic elements within the layout."""
        for item in self.content:
            item.delete()
        Widget.delete(self)

    def expand(self, width, height):
        """
        Expands to fill available vertical space.  We split available space
        equally between all spacers.
        """
        available = int((height - self.height) / len(self.expandable))
        remainder = height - self.height - len(self.expandable) * available
        for item in self.expandable:
            if remainder > 0:
                item.expand(item.width, item.height + available + 1)
                remainder -= 1
            else:
                item.expand(item.width, item.height + available)
        self.height = height
        self.width = width

    def is_expandable(self):
        """True if we contain expandable content."""
        return len(self.expandable) > 0

    def remove(self, item):
        """
        Removes a Widget from the layout.

        @param item The Widget to be removed
        """
        item.delete()
        self.content.remove(item)
        self.saved_dialog.set_needs_layout()

    def layout(self, x, y):
        """
        Lays out the child Widgets, in order from top to bottom.

        @param x X coordinate of the lower left corner
        @param y Y coordinate of the lower left corner
        """
        Widget.layout(self, x, y)

        # Expand any expandable content to our width
        for item in self.content:
            if item.is_expandable() and item.width < self.width:
                item.expand(self.width, item.height)

        top = y + self.height
        if self.align == HALIGN_RIGHT:
            for item in self.content:
                item.layout(x + self.width - item.width,
                            top - item.height)
                top -= item.height + self.padding
        elif self.align == HALIGN_CENTER:
            for item in self.content:
                item.layout(x + self.width/2 - item.width/2,
                            top - item.height)
                top -= item.height + self.padding
        else: # HALIGN_LEFT
            for item in self.content:
                item.layout(x, top - item.height)
                top -= item.height + self.padding

    def set(self, content):
        """
        Sets an entirely new set of Widgets, discarding the old.

        @param content The new list of Widgets
        """
        self.delete()
        self.content = content
        self.saved_dialog.set_needs_layout()

    def size(self, dialog):
        """
        Calculates size of the layout, based on its children.

        @param dialog The Dialog which contains the layout
        """
        if dialog is None:
            return
        Widget.size(self, dialog)
        if len(self.content) < 2:
            height = 0
        else:
            height = -self.padding
        width = 0
        for item in self.content:
            item.size(dialog)
            height += item.height + self.padding
            width = max(width, item.width)
        self.width, self.height = width, height
        self.expandable = [x for x in self.content if x.is_expandable()]

    def teardown(self):
        for item in self.content:
            item.teardown()
        self.content = []
        Widget.teardown(self)

class HorizontalLayout(VerticalLayout):
    """
    Arranges Widgets from left to right.
    """
    def __init__(self, content=[], align=VALIGN_CENTER, padding=5):
        """
        Creates a new HorizontalLayout.

        @param content A list of Widgets to be arranged
        @param align VALIGN_TOP if Widgets are to be aligned to the top
                     VALIGN_CENTER if they should be centered, and
                     VALIGN_BOTTOM if they should be aligned to the bottom.
        @param padding This amount of padding is inserted around the edge
                       of the widgets and between widgets.
        """
        VerticalLayout.__init__(self, content, align, padding)

    def expand(self, width, height):
        """
        Expands to fill available horizontal space.  We split available space
        equally between all spacers.
        """
        available = int((width - self.width) / len(self.expandable))
        remainder = height - self.height - len(self.expandable) * available
        for item in self.expandable:
            if remainder > 0:
                item.expand(item.width + available + 1, item.height)
                remainder -= 1
            else:
                item.expand(item.width + available, item.height)
        self.width = width

    def layout(self, x, y):
        """
        Lays out the child Widgets, in order from left to right.

        @param x X coordinate of the lower left corner
        @param y Y coordinate of the lower left corner
        """
        Widget.layout(self, x, y)

        # Expand any expandable content to our height
        for item in self.content:
            if item.is_expandable() and item.height < self.height:
                item.expand(item.width, self.height)

        left = x
        if self.align == VALIGN_TOP:
            for item in self.content:
                item.layout(left, y + self.height - item.height)
                left += item.width + self.padding
        elif self.align == VALIGN_CENTER:
            for item in self.content:
                item.layout(left, y + self.height/2 - item.height/2)
                left += item.width + self.padding
        else: # VALIGN_BOTTOM
            for item in self.content:
                item.layout(left, y)
                left += item.width + self.padding

    def size(self, dialog):
        """
        Calculates size of the layout, based on its children.

        @param dialog The Dialog which contains the layout
        """
        if dialog is None:
            return
        Widget.size(self, dialog)
        height = 0
        if len(self.content) < 2:
            width = 0
        else:
            width = -self.padding
        for item in self.content:
            item.size(dialog)
            height = max(height, item.height)
            width += item.width + self.padding
        self.width, self.height = width, height
        self.expandable = [x for x in self.content if x.is_expandable()]

class GridLayout(Widget):
    """
    Arranges Widgets in a table.  Each cell's height and width are set to
    the maximum width of any Widget in its column, or the maximum height of
    any Widget in its row.

    Widgets are by default aligned to the top left corner of their cells.
    Another anchor point may be specified, i.e. ANCHOR_CENTER will ensure
    that Widgets are centered within cells.
    """
    def __init__(self, content=[[]], anchor=ANCHOR_TOP_LEFT, padding=5,
                 offset=(0, 0)):
        """
        Defines a new GridLayout.

        @param content A list of rows, each of which is a list of cells
                       within that row.  'None' may be used for empty cells,
                       and rows do not need to all be the same length.
        @param anchor Alignment of
        """
        assert ((isinstance(content, list) or isinstance(content, tuple)) and
                (len(content) == 0 or (isinstance(content[0], list) or
                                       isinstance(content[0], tuple))))
        Widget.__init__(self)
        self.content = content
        self.anchor = anchor
        self.padding = padding
        self.offset = offset
        self.max_heights = []
        self.max_widths = []

    def _get_controls(self):
        """
        Returns Controls within the layout.
        """
        controls = []
        for row in self.content:
            for cell in row:
                if cell is not None:
                    controls += cell._get_controls()
        return controls

    def add_row(self, row):
        """
        Adds a new row to the layout

        @param row An array of widgets, or None for cells without widgets
        """
        assert isinstance(row, tuple) or isinstance(row, list)
        self.content.append(row)
        if self.saved_dialog is not None:
            self.saved_dialog.set_needs_layout()

    def delete(self):
        """Deletes all graphic elements within the layout."""
        for row in self.content:
            for cell in row:
                cell.delete()
        Widget.delete(self)

    def delete_row(self, row):
        """
        Deletes a row from the layout

        @param row Index of row
        """
        if len(self.content) <= row:
            return
        row = self.content.pop(row)
        for column in row:
            if column is not None:
                column.delete()
        if self.saved_dialog is not None:
            self.saved_dialog.set_needs_layout()

    def get(self, column, row):
        """
        Returns the widget located at a given column and row, or None.

        @param column Column of cell
        @param row Row of cell
        """
        if row >= len(self.content):
            return None
        row = self.content[row]
        if column >= len(row):
            return None
        else:
            return row[column]

    def layout(self, x, y):
        """
        Lays out all Widgets within this layout.

        @param x X coordinate of lower left corner
        @param y Y coordinate of lower left corner
        """
        Widget.layout(self, x, y)

        row_index = 0
        placement = Widget()
        placement.y = y + self.height
        for row in self.content:
            col_index = 0
            placement.x = x
            placement.height = self.max_heights[row_index]
            placement.y -= placement.height
            for cell in row:
                placement.width = self.max_widths[col_index]
                if cell is not None:
                    if cell.is_expandable():
                        cell.expand(placement.width, placement.height)
                    cell.layout(*GetRelativePoint(placement, self.anchor,
                                                  cell, self.anchor,
                                                  self.offset))
                placement.x += placement.width
                col_index += 1
            row_index += 1

    def set(self, column, row, item):
        """
        Sets the content of a cell within the grid.

        @param column The column of the cell to be set
        @param row The row of the cell to be set
        @param item The new Widget to be set in that cell
        """
        if len(self.content) <= row:
            self.content = list(self.content) + \
                           [] * (row - len(self.content) + 1)
        if len(self.content[row]) <= column:
            self.content[row] = list(self.content[row]) + \
                [None] * (column - len(self.content[row]) + 1)
        if self.content[row][column] is not None:
            self.content[row][column].delete()
        self.content[row][column] = item
        self.saved_dialog.set_needs_layout()

    def size(self, dialog):
        """Recalculates our size and the maximum widths and heights of
        each row and column in our table.

        @param dialog The Dialog within which we are contained
        """
        if dialog is None:
            return
        Widget.size(self, dialog)
        self.max_heights = [0] * len(self.content)
        width = 0
        for row in self.content:
            width = max(width, len(row))
        self.max_widths = [self.padding] * width
        row_index = 0
        for row in self.content:
            max_height = self.padding
            col_index = 0
            for cell in row:
                if cell is not None:
                    cell.size(dialog)
                    width, height = cell.width, cell.height
                else:
                    width = height = 0
                max_height = max(max_height, height + self.padding)
                max_width = self.max_widths[col_index]
                max_width = max(max_width, width + self.padding)
                self.max_widths[col_index] = max_width
                col_index += 1
            self.max_heights[row_index] = max_height
            row_index += 1
        if self.max_widths:
            self.width = reduce(lambda x, y: x + y, self.max_widths) \
                - self.padding
        else:
            self.width = 0
        if self.max_heights:
            self.height = reduce(lambda x, y: x + y, self.max_heights) \
                - self.padding
        else:
            self.height = 0

    def teardown(self):
        for row in self.content:
            for cell in row:
                cell.teardown()
        self.content = []
        Widget.teardown(self)

class FreeLayout(Spacer):
    """
    FreeLayout defines a rectangle on the screen where Widgets may be placed
    freely, in relation to one of its anchor points.  There is no constraints
    against the Widgets overlapping.

    FreeLayout will expand to fill available space in layouts; thus you could
    place a FreeLayout as one half of a VerticalLayout, lay out controls in
    the other half, and be assured the FreeLayout would be resized to the
    width of the overall Dialog.
    """
    def __init__(self, width=0, height=0, content=[]):
        """
        Creates a new FreeLayout.

        @param width Minimum width of FreeLayout area
        @param height Minimum height of FreeLayout area
        @param content A list of placement/Widget tuples, in the form:
                       [(ANCHOR_TOP_LEFT, 0, 0, YourWidget()),
                        (ANCHOR_TOP_RIGHT, 0, 0, YourWidget()),
                        (ANCHOR_CENTER, 30, -20, YourWidget())]
            where each tuple is (anchor, offset-x, offset-y, widget)
        """
        Spacer.__init__(self, width, height)
        self.content = content

    def _get_controls(self):
        """Returns controls within the FreeLayout"""
        controls = []
        for anchor, x, y, item in self.content:
            controls += item._get_controls()
        return controls

    def add(self, anchor, x, y, widget):
        """
        Adds a new Widget to the FreeLayout.

        @param dialog Dialog which contains the FreeLayout
        @param anchor Anchor point to set for the widget
        @param x X-coordinate of offset from anchor point; positive is to
                 the right
        @param y Y-coordinate of offset from anchor point; positive is upward
        @param widget The Widget to be added
        """
        self.content.append( (anchor, x, y, widget) )
        self.saved_dialog.set_needs_layout()

    def layout(self, x, y):
        """
        Lays out Widgets within the FreeLayout.  We make no attempt to
        assure there's enough space for them.

        @param x X coordinate of lower left corner
        @param y Y coordinate of lower left corner
        """
        Spacer.layout(self, x, y)
        for anchor, offset_x, offset_y, widget in self.content:
            x, y = GetRelativePoint(self, anchor, widget, anchor,
                                    (offset_x, offset_y))
            widget.layout(x, y)

    def remove(self, dialog, widget):
        """
        Removes a widget from the FreeLayout.

        @param dialog Dialog which contains the FreeLayout
        @param widget The Widget to be removed
        """
        self.content = [x for x in self.content if x[3] != widget]

    def size(self, dialog):
        """
        Calculate size of the FreeLayout and all Widgets inside

        @param dialog The Dialog which contains the FreeLayout
        """
        if dialog is None:
            return
        Spacer.size(self, dialog)
        for anchor, offset_x, offset_y, widget in self.content:
            widget.size(dialog)

    def teardown(self):
        for _, _, _, item in self.content:
            item.teardown()
        self.content = []
        Widget.teardown(self)

